#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Jay E. Fahlen, jay.e.fahlen@jpl.nasa.gov
#

import numpy as np
import logging

from ..core.common import eps
from ..radiative_transfer.modtran import ModtranRT
from ..radiative_transfer.six_s import SixSRT
from ..radiative_transfer.libradtran import LibRadTranRT
from ..radiative_transfer.sRTMnet import SimulatedModtranRT
from isofit.configs import Config
from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig


class RadiativeTransfer():
    """This class controls the radiative transfer component of the forward
    model. An ordered dictionary is maintained of individual RTMs (MODTRAN,
    for example). We loop over the dictionary concatenating the radiation 
    and derivatives from each RTM and interval to form the complete result.

    In general, some of the state vector components will be shared between
    RTMs and bands. For example, H20STR is shared between both VISNIR and 
    TIR. This class maintains the master list of statevectors.
    """

    def __init__(self, full_config: Config):

        # Maintain order when looping for indexing convenience
        config = full_config.forward_model.radiative_transfer

        self.statevec_names = config.statevector.get_element_names()
        self.lut_grid = config.lut_grid

        # TODO: rework this so that we instead initialize an interpolator, that calls
        # RTEs as necessary based on LUT grid or other parameters..which may happen higher up
        self.rt_engines = []
        for idx in range(len(config.radiative_transfer_engines)):
            rte_config: RadiativeTransferEngineConfig = config.radiative_transfer_engines[idx]

            if rte_config.engine_name == 'modtran':
                rte = ModtranRT(rte_config, full_config)
            elif rte_config.engine_name == 'libradtran':
                rte = LibRadTranRT(rte_config, full_config)
            elif rte_config.engine_name == '6s':
                rte = SixSRT(rte_config, full_config)
            elif rte_config.engine_name == 'simulated_modtran':
                rte = SimulatedModtranRT(rte_config, full_config)
            else:
                # Should never get here, checked in config
                raise AttributeError(
                    'Invalid radiative transfer engine name: {}'.format(rte_config.engine_name))

            self.rt_engines.append(rte)
        
        # The rest of the code relies on sorted order of the individual RT engines which cannot
        # be guaranteed by the dict JSON or YAML input
        self.rt_engines.sort(key=lambda x: x.wl[0])

        # Retrieved variables.  We establish scaling, bounds, and
        # initial guesses for each state vector element.  The state
        # vector elements are all free parameters in the RT lookup table,
        # and they all have associated dimensions in the LUT grid.
        self.bounds, self.scale, self.init = [], [], []
        self.prior_mean, self.prior_sigma = [], []

        for sv, sv_name in zip(*config.statevector.get_elements()):
            self.bounds.append(sv.bounds)
            self.scale.append(sv.scale)
            self.init.append(sv.init)
            self.prior_sigma.append(sv.prior_sigma)
            self.prior_mean.append(sv.prior_mean)

        self.bounds = np.array(self.bounds)
        self.scale = np.array(self.scale)
        self.init = np.array(self.init)
        self.prior_mean = np.array(self.prior_mean)
        self.prior_sigma = np.array(self.prior_sigma)

        self.wl = np.concatenate([RT.wl for RT in self.rt_engines])

        self.bvec = config.unknowns.get_element_names()
        self.bval = np.array([x for x in config.unknowns.get_elements()[0]])

        self.solar_irr = np.concatenate([RT.solar_irr for RT in self.rt_engines])
        # These should all be the same so just grab one
        self.coszen = [RT.coszen for RT in self.rt_engines][0]

    def xa(self):
        """Pull the priors from each of the individual RTs.
        """
        return self.prior_mean

    def Sa(self):
        """Pull the priors from each of the individual RTs.
        """
        return np.diagflat(np.power(np.array(self.prior_sigma), 2))

    def get_shared_rtm_quantities(self, x_RT, geom):
        """Return only the set of RTM quantities (transup, sphalb, etc.) that are contained
        in all RT engines.
        """

        ret = []
        for RT in self.rt_engines:
            ret.append(RT.get(x_RT, geom))

        return self.pack_arrays(ret)

    def calc_rdn(self, x_RT, rfl, Ls, geom):
        r = self.get_shared_rtm_quantities(x_RT, geom)
        L_atm = self.get_L_atm(x_RT, geom)
        L_up = Ls * r['transup']

        if geom.bg_rfl is not None:

            # adjacency effects are counted
            I = (self.solar_irr*self.coszen) / np.pi
            bg = geom.bg_rfl
            t_down = r['t_down_dif'] + r['t_down_dir']

            ret = L_atm + \
              I / (1.0-r['sphalb'] * bg) * bg * t_down * r['t_up_dif'] + \
              I / (1.0-r['sphalb'] * bg) * rfl * t_down * r['t_up_dir'] + \
              L_up

        else:
            L_down_transmitted = self.get_L_down_transmitted(x_RT, geom)
             
            ret = L_atm + \
                L_down_transmitted * rfl / (1.0 - r['sphalb'] * rfl) + \
                L_up

        return ret

    def get_L_atm(self, x_RT, geom):
        L_atms = []
        for RT in self.rt_engines:
            L_atms.append(RT.get_L_atm(x_RT, geom))
        return np.hstack(L_atms)

    def get_L_down_transmitted(self, x_RT, geom):
        L_downs = []
        for RT in self.rt_engines:
            L_downs.append(RT.get_L_down_transmitted(x_RT, geom))
        return np.hstack(L_downs)

    def drdn_dRT(self, x_RT, x_surface, rfl, drfl_dsurface, Ls,
                 dLs_dsurface, geom):

        # first the rdn at the current state vector
        rdn = self.calc_rdn(x_RT, rfl, Ls, geom)

        # perturb each element of the RT state vector (finite difference)
        K_RT = []
        x_RTs_perturb = x_RT + np.eye(len(x_RT))*eps
        for x_RT_perturb in list(x_RTs_perturb):
            rdne = self.calc_rdn(x_RT_perturb, rfl, Ls, geom)
            K_RT.append((rdne-rdn) / eps)
        K_RT = np.array(K_RT).T

        # Get K_surface
        r = self.get_shared_rtm_quantities(x_RT, geom)

        if geom.bg_rfl is not None:

             # adjacency effects are counted
            I = (self.solar_irr*self.coszen) / np.pi
            bg = geom.bg_rfl
            t_down = r['t_down_dif'] + r['t_down_dir']
            drdn_drfl = I / (1.0-r['sphalb'] * bg) * t_down * r['t_up_dir'] 

        else:
            L_down_transmitted = self.get_L_down_transmitted(x_RT, geom)
            
            # The reflected downwelling light is:
            # L_down_transmitted * rfl / (1.0 - r['sphalb'] * rfl), or
            # L_down_transmitted * rho_scaled_for_multiscattering
            # This term is the derivative of rho_scaled_for_multiscattering
            drho_scaled_for_multiscattering_drfl = 1. / (1 - r['sphalb']*rfl)**2
            
            drdn_drfl = L_down_transmitted * drho_scaled_for_multiscattering_drfl

        drdn_dLs = r['transup']
        K_surface = drdn_drfl[:, np.newaxis] * drfl_dsurface + \
            drdn_dLs[:, np.newaxis] * dLs_dsurface

        return K_RT, K_surface

    def drdn_dRTb(self, x_RT, rfl, Ls, geom):

        if len(self.bvec) == 0:
            Kb_RT = np.zeros((0, len(self.wl.shape)))

        else:
            # first the radiance at the current state vector
            r = self.get_shared_rtm_quantities(x_RT, geom)
            rdn = self.calc_rdn(x_RT, rfl, Ls, geom)

            # unknown parameters modeled as random variables per
            # Rodgers et al (2000) K_b matrix.  We calculate these derivatives
            # by finite differences
            Kb_RT = []
            perturb = (1.0+eps)
            for unknown in self.bvec:
                if unknown == 'H2O_ABSCO' and 'H2OSTR' in self.statevec_names:
                    i = self.statevec_names.index('H2OSTR')
                    x_RT_perturb = x_RT.copy()
                    x_RT_perturb[i] = x_RT[i] * perturb
                    rdne = self.calc_rdn(x_RT_perturb, rfl, Ls, geom)
                    Kb_RT.append((rdne-rdn) / eps)

        Kb_RT = np.array(Kb_RT).T
        return Kb_RT

    def summarize(self, x_RT, geom):
        ret = []
        for RT in self.rt_engines:
            ret.append(RT.summarize(x_RT, geom))
        ret = '\n'.join(ret)
        return ret

    def pack_arrays(self, rtm_quantities_from_RT_engines):
        """Take the list of dict outputs from each RT engine and
        stack their internal arrays in the same order. Keep only
        those quantities that are common to all RT engines.
        """

        # Get the intersection of the sets of keys from each of the rtm_quantities_from_RT_engines
        shared_rtm_keys = set(rtm_quantities_from_RT_engines[0].keys())
        if len(rtm_quantities_from_RT_engines) > 1:
            for rtm_quantities_from_one_RT_engine in rtm_quantities_from_RT_engines[1:]:
                shared_rtm_keys.intersection_update(rtm_quantities_from_one_RT_engine.keys())

        # Concatenate the different band ranges
        rtm_quantities_concatenated_over_RT_bands = {}
        for key in shared_rtm_keys:
            temp = [x[key] for x in rtm_quantities_from_RT_engines]
            rtm_quantities_concatenated_over_RT_bands[key] = np.hstack(temp)

        return rtm_quantities_concatenated_over_RT_bands



#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
import scipy
from sklearn.cluster import KMeans
from spectral.io import envi
import os

from isofit.core.common import expand_path, json_load_ascii


def surface_model(config_path: str, wavelength_path: str = None, 
        output_path: str = None) -> None:
    """The surface model tool contains everything you need to build basic
    multicomponent (i.e. colleciton of Gaussian) surface priors for the
    multicomponent surface model.

    Args:
        config_path: path to a JSON formatted surface model configuration
        wavelength_path: optional path to a three-column wavelength file, 
           overriding the configuration file settings
        output_path: optional path to the destination .mat file, overriding
           the configuration file settings
    Returns:
        None
    """

    # Load configuration JSON into a local dictionary
    configdir, _ = os.path.split(os.path.abspath(config_path))
    config = json_load_ascii(config_path, shell_replace=True)

    # Determine top level parameters
    for q in ['output_model_file', 'sources', 'normalize', 'wavelength_file']:
        if q not in config:
            raise ValueError("Missing parameter: %s" % q)
    if wavelength_path is not None:
        wavelength_file = wavelength_path
    else:
        wavelength_file = expand_path(configdir, config['wavelength_file'])
    if output_path is not None:
        outfile = output_path
    else:
        outfile = expand_path(configdir, config['output_model_file'])
    normalize = config['normalize']
    reference_windows = config['reference_windows']

    # load wavelengths file, and change units to nm if needed
    q = np.loadtxt(wavelength_file)
    if q.shape[1] > 2:
        q = q[:, 1:]
    if q[0, 0] < 100:
        q = q * 1000.0
    wl = q[:, 0]
    nchan = len(wl)

    # build global reference windows
    refwl = []
    for wi, window in enumerate(reference_windows):
        active_wl = np.logical_and(wl >= window[0], wl < window[1])
        refwl.extend(wl[active_wl])
    normind = np.array([np.argmin(abs(wl - w)) for w in refwl])
    refwl = np.array(refwl, dtype=float)

    # create basic model template
    model = {
        'normalize': normalize,
        'wl': wl,
        'means': [],
        'covs': [],
        'attribute_means': [],
        'attribute_covs': [],
        'attributes': [],
        'refwl': refwl
    }

    # each "source" (i.e. spectral library) is treated separately
    for si, source_config in enumerate(config['sources']):

        # Determine source parameters
        for q in ['input_spectrum_files', 'windows', 'n_components', 'windows']:
            if q not in source_config:
                raise ValueError(
                    'Source %i is missing a parameter: %s' % (si, q))

        # Determine whether we should synthesize our own mixtures
        if 'mixtures' in source_config:
            mixtures = source_config['mixtures']
        elif 'mixtures' in config:
            mixtures = config['mixtures']
        else:
            mixtures = 0

        # open input files associated with this source
        infiles = [expand_path(configdir, fi) for fi in
                   source_config['input_spectrum_files']]

        # associate attributes, if they exist. These will not be used
        # in the retrieval, but can be used in post-analysis
        if 'input_attribute_files' in source_config:
            infiles_attributes = [expand_path(configdir, fi) for fi in
                   source_config['input_attribute_files']]
            if len(infiles_attributes) != len(infiles):
                raise IndexError('spectrum / attribute file mismatch')
        else:
            infiles_attributes = [None for fi in 
                source_config['input_spectrum_files']]

        ncomp = int(source_config['n_components'])
        windows = source_config['windows']

        # load spectra
        spectra, attributes = [],[]
        for infile, attribute_file in zip(infiles, infiles_attributes):

            hdrfile = infile + '.hdr'
            rfl = envi.open(hdrfile, infile)
            nl, nb, ns = [int(rfl.metadata[n])
                          for n in ('lines', 'bands', 'samples')]
            swl = np.array([float(f) for f in rfl.metadata['wavelength']])

            # Maybe convert to nanometers
            if swl[0] < 100:
                swl = swl * 1000.0

            # Load library and adjust interleave, if needed
            rfl_mm = rfl.open_memmap(interleave='bip', writable=False)
            x = np.array(rfl_mm[:, :, :])
            x = x.reshape(nl * ns, nb)

            # import spectra and resample
            for x1 in x:
                p = scipy.interpolate.interp1d(swl, x1, kind='linear', bounds_error=False,
                             fill_value='extrapolate')
                spectra.append(p(wl))

            # Load attributes
            if attribute_file is not None:

                hdrfile = attribute_file + '.hdr'
                attr = envi.open(hdrfile, attribute_file)
                nla, nba, nsa = [int(attr.metadata[n])
                          for n in ('lines', 'bands', 'samples')]

                # Load library and adjust interleave, if needed
                attr_mm = attr.open_memmap(interleave='bip', writable=False)
                x = np.array(attr_mm[:, :, :])
                x = x.reshape(nla * nsa, nba)
                model['attributes'] = attr.metadata['band names']

                # import spectra and resample
                for x1 in x:
                    attributes.append(x1)

        if len(attributes)>0 and len(attributes) != len(spectra):
            raise IndexError('Mismatch in number of spectra vs. attributes')


        # calculate mixtures, if needed
        if len(attributes)>0 and mixtures > 0:
            raise ValueError('Synthetic mixtures w/ attributes is not advised')

        n = float(len(spectra))
        nmix = int(n * mixtures)
        for mi in range(nmix):
            s1, m1 = spectra[int(np.random.rand() * n)], np.random.rand()
            s2, m2 = spectra[int(np.random.rand() * n)], 1.0 - m1
            spectra.append(m1 * s1 + m2 * s2)

        # Lists to arrays
        spectra = np.array(spectra)
        attributes = np.array(attributes)

        # Flag bad data
        use = np.all(np.isfinite(spectra), axis=1)
        spectra = spectra[use, :]
        if len(attributes)>0:
            attributes = attributes[use,:]

        # Accumulate total list of window indices
        window_idx = -np.ones((nchan), dtype=int)
        for wi, win in enumerate(windows):
            active_wl = np.logical_and(wl >= win['interval'][0], wl < win['interval'][1])
            window_idx[active_wl] = wi

        # Two step model generation.  First step is k-means clustering.
        # This is more "stable" than Expectation Maximization with an 
        # unconstrained covariance matrix    
        kmeans = KMeans(init='k-means++', n_clusters=ncomp, n_init=10)
        kmeans.fit(spectra)
        Z = kmeans.predict(spectra)

        # Build a combined dataset of attributes and spectra
        if len(attributes)>0:
            spectra_attr = np.concatenate((spectra,attributes), axis=1)

        # Now fit the full covariance for each component
        for ci in range(ncomp):

            m = np.mean(spectra[Z == ci, :], axis=0)
            C = np.cov(spectra[Z == ci, :], rowvar=False)
            if len(attributes) > 0:
                m_attr = np.mean(spectra_attr[Z == ci, :], axis=0)
                C_attr = np.cov(spectra_attr[Z == ci, :], rowvar=False)

            for i in range(nchan):
                window = windows[window_idx[i]]

                # Each spectral interval, or window, is constructed
                # using one of several rules.  We can draw the covariance
                # directly from the data...
                if window['correlation'] == 'EM':
                    C[i, i] = C[i, i] + float(window['regularizer'])

                # Alternatively, we can use a band diagonal form,
                # a Gaussian process that promotes local smoothness.
                elif window['correlation'] == 'GP':
                    width = float(window['gp_width'])
                    magnitude = float(window['gp_magnitude'])
                    kernel = scipy.stats.norm.pdf((wl-wl[i])/width)
                    kernel = kernel/kernel.sum() * magnitude
                    C[i, :] = kernel
                    C[:, i] = kernel
                    C[i, i] = C[i, i] + float(window['regularizer'])

                # To minimize bias, leave the channels independent
                # and uncorrelated
                elif window['correlation'] == 'decorrelated':
                    ci = C[i, i]
                    C[:, i] = 0
                    C[i, :] = 0
                    C[i, i] = ci + float(window['regularizer'])

                else:
                    raise ValueError(
                        'I do not recognize the method ' + window['correlation'])

            # Normalize the component spectrum if desired
            if normalize == 'Euclidean':
                z = np.sqrt(np.sum(pow(m[normind], 2)))
            elif normalize == 'RMS':
                z = np.sqrt(np.mean(pow(m[normind], 2)))
            elif normalize == 'None':
                z = 1.0
            else:
                raise ValueError(
                    'Unrecognized normalization: %s\n' % normalize)
            m = m / z
            C = C / (z ** 2)

            model['means'].append(m)
            model['covs'].append(C)

            if len(attributes)>0:
                model['attribute_means'].append(m_attr)
                model['attribute_covs'].append(C_attr)

    model['means'] = np.array(model['means'])
    model['covs'] = np.array(model['covs'])
    model['attribute_means'] = np.array(model['attribute_means'])
    model['attribute_covs'] = np.array(model['attribute_covs'])

    scipy.io.savemat(outfile, model)



# David R Thompson, Adam Erickson

from isofit.core.isofit import Isofit
from isofit.utils import surface_model


# Build the surface model
surface_model("configs/ang20171108t184227_surface.json")

# Run retrievals
model = Isofit("configs/ang20171108t184227_beckmanlawn-libradtran.json")
model.run()


#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

from os.path import split, abspath
import numpy as np
import scipy
from scipy.ndimage.filters import gaussian_filter1d
from spectral.io import envi

from isofit.core.common import expand_path, json_load_ascii


def instrument_model(config):
    """."""

    hdr_template = """ENVI
    samples = {samples}
    lines   = {lines}
    bands   = 1
    header offset = 0
    file type = ENVI Standard
    data type = 4
    interleave = bsq
    byte order = 0
    """

    config = json_load_ascii(config, shell_replace=True)
    configdir, configfile = split(abspath(config))

    infile = expand_path(configdir, config['input_radiance_file'])
    outfile = expand_path(configdir, config['output_model_file'])
    flatfile = expand_path(configdir, config['output_flatfield_file'])
    uniformity_thresh = float(config['uniformity_threshold'])

    infile_hdr = infile + '.hdr'
    img = envi.open(infile_hdr, infile)
    inmm = img.open_memmap(interleave='bil', writable=False)
    X = np.array(inmm[:, :, :], dtype=np.float32)
    nr, nb, nc = X.shape

    FF, Xhoriz, Xhorizp, use_ff = _flat_field(X, uniformity_thresh)
    np.array(FF, dtype=np.float32).tofile(flatfile)
    with open(flatfile+'.hdr', 'w') as fout:
        fout.write(hdr_template.format(lines=nb, samples=nc))

    C, Xvert, Xvertp, use_C = _column_covariances(X, uniformity_thresh)
    cshape = (C.shape[0], C.shape[1]**2)
    out = np.array(C, dtype=np.float32).reshape(cshape)
    mdict = {'columns': out.shape[0], 'bands': out.shape[1],
             'covariances': out, 'Xvert': Xvert, 'Xhoriz': Xhoriz,
             'Xvertp': Xvertp, 'Xhorizp': Xhorizp, 'use_ff': use_ff,
             'use_C': use_C}
    scipy.io.savemat(outfile, mdict)



def _high_frequency_vert(X, sigma=4.0):
    """."""

    nl, nb, nr = X.shape
    Xvert = X.copy()
    for r in range(nr):
        for b in range(nb):
            filt = gaussian_filter1d(Xvert[:, b, r], sigma, mode='nearest')
            Xvert[:, b, r] = X[:, b, r] - filt
    return Xvert


def _low_frequency_horiz(X, sigma=4.0):
    """."""

    nl, nb, nr = X.shape
    Xhoriz = X.copy()
    for l in range(nl):
        for b in range(nb):
            Xhoriz[l, b, :] = gaussian_filter1d(
                Xhoriz[l, b, :], sigma, mode='nearest')
    return Xhoriz


def _flat_field(X, uniformity_thresh):
    """."""

    Xhoriz = _low_frequency_horiz(X, sigma=4.0)
    Xhorizp = _low_frequency_horiz(X, sigma=3.0)
    nl, nb, nc = X.shape
    FF = np.zeros((nb, nc))
    use_ff = np.ones((X.shape[0], X.shape[2])) > 0
    for b in range(nb):
        xsub = Xhoriz[:, b, :]
        mu = xsub.mean(axis=0)
        dists = abs(xsub - mu)
        thresh = np.percentile(dists.flatten(), 90.0)
        use = dists < thresh
        FF[b, :] = ((xsub*use).sum(axis=0)/use.sum(axis=0)) / \
            ((X[:, b, :]*use).sum(axis=0)/use.sum(axis=0))
        use_ff = np.logical_and(use_ff, use)
    return FF, Xhoriz, Xhorizp, np.array(use_ff)


def _column_covariances(X, uniformity_thresh):
    """."""

    Xvert = _high_frequency_vert(X, sigma=4.0)
    Xvertp = _high_frequency_vert(X, sigma=3.0)
    models = []
    use_C = []
    for i in range(X.shape[2]):
        xsub = Xvert[:, :, i]
        mu = xsub.mean(axis=0)
        dists = np.sqrt(pow((xsub - mu), 2).sum(axis=1))
        thresh = np.percentile(dists, 95.0)
        use = dists < thresh
        C = np.cov(xsub[use, :], rowvar=False)
        [U, V, D] = scipy.linalg.svd(C)
        V[V < 1e-8] = 1e-8
        C = U.dot(np.diagflat(V)).dot(D)
        models.append(C)
        use_C.append(use)
    return np.array(models), Xvert, Xvertp, np.array(use_C).T



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: Philip G. Brodrick, philip.brodrick@jpl.nasa.gov

from typing import Dict, List, Type
from isofit.configs.base_config import BaseConfigSection
from isofit.configs.sections.inversion_config import InversionConfig
import os


class ImplementationConfig(BaseConfigSection):

    def __init__(self, sub_configdic: dict = None):
        """
        Input file(s) configuration.
        """

        self._mode_type = str
        self.mode = 'inversion'
        """
        str: Defines the operating mode for isofit. Current options are: inversion, inversion_mcmc, 
        and 'simulation'.
        """

        self._inversion_type = InversionConfig
        self.inversion: InversionConfig = None
        """InversionConfig: optional config for running in inversion mode."""

        self._n_cores_type = int
        self.n_cores = None
        """int: number of cores to use."""

        self._task_inflation_factor_type = int
        self.task_inflation_factor = 10
        """int: Submit task_inflation_factor*n_cores number of tasks."""

        self._ip_head_type = str
        self.ip_head = None
        """str: Ray - parameter.  IP-head (for multi-node runs)."""

        self._redis_password_type = str
        self.redis_password = None
        """str: Ray - parameter.  Redis-password (for multi-node runs)."""

        self._rte_configure_and_exit_type = bool
        self.rte_configure_and_exit = False
        """bool: Indicates that code should terminate as soon as all radiative transfer engine configuration files are
        written (without running them)"""

        self._rte_auto_rebuild_type = bool
        self.rte_auto_rebuild = True
        """bool: Flag indicating whether radiative transfer engines should automatically rebuild."""

        self._ray_temp_dir_type = str
        self.ray_temp_dir = '/tmp/ray'
        """str: Overrides the standard ray temporary directory.  Useful for multiuser systems."""

        self._ray_ignore_reinit_error_type = bool
        self.ray_ignore_reinit_error = True
        """bool: Boolean to tell ray to ignore re-initilaization.  Can be convenient for multiple Isofit instances."""

        self.set_config_options(sub_configdic)

    def _check_config_validity(self) -> List[str]:
        errors = list()

        valid_implementation_modes = ['inversion', 'mcmc_inversion', 'simulation']
        if self.mode not in valid_implementation_modes:
            errors.append('Invalid implementation mode: {}.  Valid options are: {}'.
                          format(self.mode, valid_implementation_modes))

        if self.mode != 'simulation' and self.inversion is None:
            errors.append('If running outside of simulation mode, Inversion must be defined')
        elif self.mode == 'simulation' and self.inversion is None:
            #TODO: fix this
            errors.append('Even in simulation mode, and inversion config must be defined, though'
                          'it may be blank.')

        if int(self.ip_head is not None) + int(self.redis_password is not None) == 1:
            errors.append('If either ip_head or redis_password are specified, both must be specified')

        return errors



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
import scipy.linalg

from ..core.common import emissive_radiance, eps
from .surface_multicomp import MultiComponentSurface
from isofit.configs import Config


class ThermalSurface(MultiComponentSurface):
    """A model of the surface based on a Mixture of a hot Black Body and 
        Multicomponent cold surfaces."""

    def __init__(self, full_config: Config):
        """."""

        config = full_config.forward_model.surface

        super().__init__(full_config)

        # TODO: Enforce this attribute in the config, not here (this is hidden)
        # Handle additional state vector elements
        self.statevec_names.extend(['SURF_TEMP_K'])
        self.init.extend([300.0])  # This is overwritten below
        self.scale.extend([100.0])
        self.bounds.extend([[250.0, 400.0]])
        self.surf_temp_ind = len(self.statevec_names) - 1
        self.emissive = True
        self.n_state = len(self.init)

        self.emissivity_for_surface_T_init = \
                config.emissivity_for_surface_T_init
        self.surface_T_prior_sigma_degK = config.surface_T_prior_sigma_degK

    def xa(self, x_surface, geom):
        """Mean of prior distribution, calculated at state x.  We find
        the covariance in a normalized space (normalizing by z) and then un-
        normalize the result for the calling function."""

        mu = MultiComponentSurface.xa(self, x_surface, geom)
        mu[self.surf_temp_ind] = self.init[self.surf_temp_ind]
        return mu

    def Sa(self, x_surface, geom):
        """Covariance of prior distribution, calculated at state x."""

        Cov = MultiComponentSurface.Sa(self, x_surface, geom)
        Cov[self.surf_temp_ind, self.surf_temp_ind] = \
            self.surface_T_prior_sigma_degK**2

        return Cov

    def fit_params(self, rfl_meas, geom, *args):
        """Given a reflectance estimate, find the surface reflectance"""

        x_surface = MultiComponentSurface.fit_params(self, rfl_meas, geom)
        x_surface[self.surf_temp_ind] = self.init[self.surf_temp_ind]

        return x_surface

    def calc_rfl(self, x_surface, geom):
        """Reflectance. This could be overriden to add (for example)
            specular components"""

        return self.calc_lamb(x_surface, geom)

    def drfl_dsurface(self, x_surface, geom):
        """Partial derivative of reflectance with respect to state vector, 
        calculated at x_surface."""

        return self.dlamb_dsurface(x_surface, geom)

    def calc_lamb(self, x_surface, geom):
        """Lambertian Reflectance."""

        return MultiComponentSurface.calc_lamb(self, x_surface, geom)

    def dlamb_dsurface(self, x_surface, geom):
        """Partial derivative of Lambertian reflectance with respect to state 
        vector, calculated at x_surface."""

        dlamb = MultiComponentSurface.dlamb_dsurface(self, x_surface, geom)
        dlamb[:, self.surf_temp_ind] = 0
        return dlamb

    def calc_Ls(self, x_surface, geom):
        """Emission of surface, as a radiance."""

        T = x_surface[self.surf_temp_ind]
        rfl = self.calc_rfl(x_surface, geom)
        rfl[rfl > 1.] = 1.
        emissivity = 1 - rfl
        Ls, dLs_dT = emissive_radiance(emissivity, T, self.wl)
        return Ls

    def dLs_dsurface(self, x_surface, geom):
        """Partial derivative of surface emission with respect to state vector, 
        calculated at x_surface."""

        T = x_surface[self.surf_temp_ind]
        lambertian_rfl = self.calc_lamb(x_surface, geom)
        emissivity = 1 - lambertian_rfl
        Ls, dLs_dT = emissive_radiance(emissivity, T, self.wl)
        dLs_drfl = np.diag(-1*Ls)
        dLs_dsurface = np.vstack([dLs_drfl, dLs_dT]).T

        return dLs_dsurface

    def summarize(self, x_surface, geom):
        """Summary of state vector."""

        mcm = MultiComponentSurface.summarize(self, x_surface, geom)
        msg = ' Kelvins: %5.1f ' % x_surface[self.surf_temp_ind]
        return msg+mcm



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import logging
from datetime import datetime
import numpy as np

from .sunposition import sunpos


class Geometry:
    """The geometry of the observation, all we need to calculate sensor,
      surface, and solar positions."""

    def __init__(self, obs=None, glt=None, loc=None, ds=None,
                 esd=None, pushbroom_column=None, bg_rfl=None):

        # Set some benign defaults...
        self.earth_sun_file = None
        self.observer_zenith = 0
        self.observer_azimuth = 0
        self.observer_altitude_km = None
        self.surface_elevation_km = None
        self.datetime = None
        self.day_of_year = None
        self.latitude = None
        self.longitude = None
        self.longitudeE = None
        self.gmtime = None
        self.earth_sun_distance = None
        self.OBSZEN = 180.0
        self.RELAZ = 0.0
        self.TRUEAZ = 0.0
        self.H1ALT = None
        self.umu = 1.0
        self.pushbroom_column = pushbroom_column
        self.bg_rfl = bg_rfl

        # The 'obs' object is observation metadata that follows a historical
        # AVIRIS-NG format.  It arrives to our initializer in the form of
        # a list-like object...
        if obs is not None:
            self.path_length = obs[0]
            self.observer_azimuth = obs[1]  # 0 to 360 clockwise from N
            self.observer_zenith = obs[2]  # 0 to 90 from zenith
            self.solar_azimuth = obs[3]  # 0 to 360 clockwise from N
            self.solar_zenith = obs[4]  # 0 to 90 from zenith
            self.OBSZEN = 180.0 - abs(obs[2])  # MODTRAN convention?
            self.RELAZ = obs[1] - obs[3] + 180.0
            self.TRUEAZ = obs[1]  # MODTRAN convention?
            self.umu = np.cos(obs[2]/360.0*2.0*np.pi)  # Libradtran

        # The 'loc' object is a list-like object that optionally contains
        # latitude and longitude information about the surface being
        # observed.
        if loc is not None:
            self.surface_elevation_km = loc[2] / 1000.0
            self.latitude = loc[1]
            self.longitude = loc[0]
            self.longitudeE = -loc[0]
            if self.longitude < 0:
                self.longitude = 360.0 - self.longitude

            logging.debug('Geometry lat: %f lon: %f' %
                          (self.latitude, self.longitude))
            logging.debug('Geometry observer OBSZEN: %f RELAZ: %f GNDALT: %f' %
                          (self.OBSZEN, self.RELAZ, self.surface_elevation_km))
        
        if loc is not None and obs is not None:
            self.H1ALT = self.surface_elevation_km + self.path_length*np.cos(np.deg2rad(self.observer_zenith))
            self.observer_altitude_km = self.surface_elevation_km + self.path_length*np.cos(np.deg2rad(self.observer_zenith))

        # The ds object is an optional date object, defining the time of
        # the observation.
        if ds is not None:
            self.datetime = datetime.strptime(ds, '%Y%m%dt%H%M%S')
            self.day_of_year = self.datetime.timetuple().tm_yday

        # Finally, the earth sun distance is an array that maps the day of the
        # year (zero-indexed!) onto the mean-relative distance to the sun.
        if esd is not None:
            self.earth_sun_distance = esd.copy()

    def coszen(self):
        """ Return the cosine of the solar zenith."""
        self.dt = self.datetime
        az, zen, ra, dec, h = sunpos(self.datetime, self.latitude,
                                     self.longitudeE,
                                     self.surface_elevation_km * 1000.0,
                                     radians=True)
        return np.cos(zen)

    def sundist(self):
        '''Return the mean-relative distance to the sun as defined by the
        day of the year.  Note that we use zero-indexed table, offset by one 
        from the actual cardenality, per Python conventions...'''
        return float(self.earth_sun_distance[self.day_of_year-1, 1])



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import sys
import pep8
from glob import glob
from os.path import split, abspath


testdir, fname = split(abspath(__file__))
config_file = testdir+'/data/pep8_config.txt'
excludes = ['sunposition.py']


def test_pep8_conformance():
    """Test that we conform to PEP8."""

    config_file = 'data/pep8_config.txt'
    files = []
    for fi in glob('../*.py'):
        files.append(fi)

    # After acceptance, this will be uncommented
    pep8style = pep8.StyleGuide(config_file=config_file, quiet=False)
    result = pep8style.check_files(files)
    if result.total_errors != 0:
        print("Found PEP8 conformance error.")
        print("Please fix your style with autopep8.")
    assert result.total_errors == 0



#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import os
import logging
import numpy as np
from scipy.interpolate import interp1d

from ..core.common import resample_spectrum, VectorInterpolator
from .look_up_tables import TabularRT, FileExistsError

from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig
from isofit.configs import Config


### Variables ###

eps = 1e-5  # used for finite difference derivative calculations

### Classes ###


class LibRadTranRT(TabularRT):
    """A model of photon transport including the atmosphere."""

    def __init__(self, engine_config: RadiativeTransferEngineConfig, full_config: Config):

        self.angular_lut_keys_degrees = ['OBSZEN', 'TRUEAZ', 'viewzen', 'viewaz',
                                         'solzen', 'solaz']
        self.angular_lut_keys_radians = []

        super().__init__(engine_config, full_config)
        self.treat_as_emissive = False
        self.libradtran_dir = self.find_basedir(engine_config)
        self.libradtran_template_file = engine_config.template_file
        self.libradtran_environment = engine_config.environment

        self.lut_quantities = ['rhoatm', 'transm', 'sphalb', 'transup']

        # Build the lookup table
        self.build_lut()

    def find_basedir(self, config: RadiativeTransferEngineConfig):
        """Seek out a libradtran base directory."""

        if config.engine_base_dir is not None:
            return config.engine_base_dir

        try:
            return os.getenv('LIBRADTRAN_DIR')
        except KeyError:
            pass
        return None

    def rebuild_cmd(self, point, fn):
        """."""

        # start with defaults
        vals = {'atmosphere': 'midlatitude_summer'}
        for n, v in zip(self.lut_names, point):
            vals[n] = v

        # Translate a couple of special cases
        if 'AOT550' in self.lut_names:
            vals['aerosol_visibility'] = self.ext550_to_vis(vals['AOT550'])
        if 'H2OSTR' in self.lut_names:
            vals['h2o_mm'] = vals['H2OSTR']*10.0

        with open(self.libradtran_template_file, 'r') as fin:
            template = fin.read()
            dict0, dict025, dict05 = [dict(vals).copy() for q in (1, 2, 3)]
            dict0['albedo'] = '0.0'
            dict025['albedo'] = '0.25'
            dict05['albedo'] = '0.5'
            libradtran_config_str0 = template.format(**dict0)
            libradtran_config_str025 = template.format(**dict025)
            libradtran_config_str05 = template.format(**dict05)

        # Check rebuild conditions: LUT is missing or from a different config
        infilename0 = 'LUT_'+fn+'_alb0.inp'
        infilename05 = 'LUT_'+fn+'_alb05.inp'
        infilename025 = 'LUT_'+fn+'_alb025.inp'
        infilepath0 = os.path.join(self.lut_dir, infilename0)
        infilepath05 = os.path.join(self.lut_dir, infilename05)
        infilepath025 = os.path.join(self.lut_dir, infilename025)

        outfilename0 = 'LUT_'+fn+'_alb0.out'
        outfilename05 = 'LUT_'+fn+'_alb05.out'
        outfilename025 = 'LUT_'+fn+'_alb025.out'
        outfilenamezen = 'LUT_'+fn+'.zen'
        outfilepath0 = os.path.join(self.lut_dir, outfilename0)
        outfilepath05 = os.path.join(self.lut_dir, outfilename05)
        outfilepath025 = os.path.join(self.lut_dir, outfilename025)
        outfilepathzen = os.path.join(self.lut_dir, outfilenamezen)

        scriptfilename = 'LUT_'+fn+'.sh'
        scriptfilepath = os.path.join(self.lut_dir, scriptfilename)

        # Are all files present?
        rebuild = False
        for path in [infilepath0, infilepath05, infilepath025,
                     outfilepath0, outfilepath05, outfilepath025,
                     outfilepathzen, scriptfilepath]:
            if not os.path.exists(path):
                rebuild = True

        # Has configuration changed?
        if not rebuild:
            current0 = open(infilepath0, 'r').read()
            current05 = open(infilepath05, 'r').read()
            current025 = open(infilepath025, 'r').read()
            rebuild = (rebuild or (libradtran_config_str0 != current0))
            rebuild = (rebuild or (libradtran_config_str025 != current025))
            rebuild = (rebuild or (libradtran_config_str05 != current05))

        if not rebuild:
            raise FileExistsError('Files exist')

        if self.libradtran_dir is None:
            logging.error('Specify a LibRadTran installation')
            raise KeyError('Specify a LibRadTran installation')

        # write config files
        with open(infilepath0, 'w') as f:
            f.write(libradtran_config_str0)
        with open(infilepath025, 'w') as f:
            f.write(libradtran_config_str025)
        with open(infilepath05, 'w') as f:
            f.write(libradtran_config_str05)

        # Find the location and time for solar zenith caching
        with open(infilepath0, 'r') as fin:
            lat, lon, yr, mon, day, hour, mn = \
                None, None, None, None, None, None, None
            for line in fin.readlines():
                # Skip comment lines
                if line.strip().startswith("#"):
                    continue
                if 'latitude N' in line:
                    lat = float(line.split()[-1])
                elif 'latitude S' in line:
                    lat = -float(line.split()[-1])
                elif 'longitude W' in line:
                    lon = float(line.split()[-1])
                elif 'longitude E' in line:
                    lon = -float(line.split()[-1])
                elif 'time' in line:
                    yr, mon, day, hour, mn, sec = [
                        float(q) for q in line.split()[1:]]

        # Write runscript file
        with open(scriptfilepath, 'w') as f:
            f.write('#!/usr/bin/bash\n')
            f.write('%s\n' % self.libradtran_environment)
            f.write('(cd %s/bin && ./uvspec < %s > %s)\n' %
                    (self.libradtran_dir, infilepath0, outfilepath0))
            f.write('(cd %s/bin && ./uvspec < %s > %s)\n' %
                    (self.libradtran_dir, infilepath05, outfilepath05))
            f.write('(cd %s/bin && ./uvspec < %s > %s)\n' %
                    (self.libradtran_dir, infilepath025, outfilepath025))
            f.write('(cd %s/bin && ./zenith %s -a %s -o %s -y %s %s %s %s %s > %s)\n' %
                    (self.libradtran_dir,
                     '-s 0 -q', lat, lon, yr, day, mon, hour, mn,
                     outfilepathzen))

        return 'bash '+scriptfilepath

    def load_rt(self, fn):
        """Load the results of a LibRadTran run."""

        _, rdn0,   _ = np.loadtxt(self.lut_dir+'/LUT_'+fn+'_alb0.out').T
        _, rdn025, _ = np.loadtxt(self.lut_dir+'/LUT_'+fn+'_alb025.out').T
        wl, rdn05,  irr = np.loadtxt(self.lut_dir+'/LUT_'+fn+'_alb05.out').T

        # Replace a few zeros in the irradiance spectrum via interpolation
        good = irr > 1e-15
        bad = np.logical_not(good)
        irr[bad] = interp1d(wl[good], irr[good])(wl[bad])

        # Translate to Top of Atmosphere (TOA) reflectance
        rhoatm = rdn0 / 10.0 / irr * np.pi  # Translate to uW nm-1 cm-2 sr-1
        rho025 = rdn025 / 10.0 / irr * np.pi
        rho05 = rdn05 / 10.0 / irr * np.pi

        # Resample TOA reflectances to simulate the instrument observation
        rhoatm = resample_spectrum(rhoatm, wl, self.wl, self.fwhm)
        rho025 = resample_spectrum(rho025, wl, self.wl, self.fwhm)
        rho05 = resample_spectrum(rho05,  wl, self.wl, self.fwhm)
        irr = resample_spectrum(irr,    wl, self.wl, self.fwhm)

        # Calculate some atmospheric optical constants NOTE: This calc is not
        # numerically stable for cases where rho025 and rho05 are the same.
        # Anecdotally, in all of these cases, they are also the same as rhoatm,
        # so the equation reduces to 0 / 0. Therefore, we assume that spherical
        # albedo here is zero. Any other non-finite results are (currently)
        # unexpected, so we convert them to errors.
        bad = np.logical_and(rho025 == rhoatm, rho05 == rhoatm)
        sphalb = 2.8*(2.0*rho025-rhoatm-rho05)/(rho025-rho05)
        if np.sum(bad) > 0:
            logging.debug('Setting sphalb = 0 where rho025 == rho05 == rhoatm.')
            sphalb[bad] = 0

        if not np.all(np.isfinite(sphalb)):
            raise AttributeError('Non-finite values in spherical albedo calculation')

        transm = (rho05-rhoatm)*(2.0-sphalb)

        # For now, don't estimate this term!!
        # TODO: Have LibRadTran calculate it directly
        transup = np.zeros(self.wl.shape)

        # Get solar zenith, translate to irradiance at zenith = 0
        # HACK: If a file called `prescribed_geom` exists in the LUT directory,
        # use that instead of the LibRadtran calculated zenith angle. This is
        # not the most elegant or efficient solution, but it seems to work.
        zenfile = os.path.join(self.lut_dir, "prescribed_geom")
        if not os.path.exists(zenfile):
            zenfile = os.path.join(self.lut_dir, 'LUT_'+fn+'.zen')
        with open(zenfile, 'r') as fin:
            output = fin.read().split()
            solzen, solaz = [float(q) for q in output[1:]]

        self.coszen = np.cos(solzen/360.0*2.0*np.pi)
        irr = irr / self.coszen
        self.solar_irr = irr.copy()

        results = {"wl": self.wl, 'solzen': solzen, 'irr': irr,
                   "solzen": solzen, "rhoatm": rhoatm, "transm": transm,
                   "sphalb": sphalb, "transup": transup}
        return results

    def ext550_to_vis(self, ext550):
        return np.log(50.0) / (ext550 + 0.01159)

    def build_lut(self, rebuild=False):

        TabularRT.build_lut(self, rebuild)

        librt_outputs = []
        for point, fn in zip(self.points, self.files):
            librt_outputs.append(self.load_rt(fn))

        self.cache = {}
        dims_aug = self.lut_dims + [self.n_chan]
        for key in self.lut_quantities:
            temp = np.zeros(dims_aug, dtype=float)
            for librt_output, point in zip(librt_outputs, self.points):
                ind = [np.where(g == p)[0] for g, p in
                       zip(self.lut_grids, point)]
                ind = tuple(ind)
                temp[ind] = librt_output[key]

            self.luts[key] = VectorInterpolator(self.lut_grids, temp,
                                                self.lut_interp_types)

    def _lookup_lut(self, point):
        ret = {}
        for key, lut in self.luts.items():
            ret[key] = np.array(lut(point)).ravel()
        return ret

    def get(self, x_RT, geom):
        point = np.zeros((self.n_point,))
        for point_ind, name in enumerate(self.lut_grid_config):
            if name in self.statevector_names:
                ix = self.statevector_names.index(name)
                point[point_ind] = x_RT[ix]
            elif name == "OBSZEN":
                point[point_ind] = geom.OBSZEN
            elif name == "GNDALT":
                point[point_ind] = geom.GNDALT
            elif name == "viewzen":
                point[point_ind] = geom.observer_zenith
            elif name == "viewaz":
                point[point_ind] = geom.observer_azimuth
            elif name == "solaz":
                point[point_ind] = geom.solar_azimuth
            elif name == "solzen":
                point[point_ind] = geom.solar_zenith
            elif name == "TRUEAZ":
                point[point_ind] = geom.TRUEAZ
            elif name == 'phi':
                point[point_ind] = geom.phi
            elif name == 'umu':
                point[point_ind] = geom.umu
            else:
                # If a variable is defined in the lookup table but not
                # specified elsewhere, we will default to the minimum
                point[point_ind] = min(self.lut_grid_config[name])
        return self._lookup_lut(point)

    def get_L_atm(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rho = r['rhoatm']
        rdn = rho / np.pi*(self.solar_irr * self.coszen)
        return rdn

    def get_L_down_transmitted(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rdn = (self.solar_irr * self.coszen) / np.pi * r['transm']
        return rdn



#! /usr/bin/env python3
#
# ISOFIT redistributes this version of sunposition.py for ease of use and
# and compatibility under the terms of The MIT License (MIT):
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Samuel Bear Powell
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
from datetime import datetime


class _sp:
    """."""

    @staticmethod
    def calendar_time(dt):
        """."""

        try:
            x = dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond
            return x
        except AttributeError:
            try:
                # will raise OSError if dt is not acceptable
                return _sp.calendar_time(datetime.utcfromtimestamp(dt))
            except BaseException:
                raise TypeError(
                    'dt must be datetime object or POSIX timestamp')

    @staticmethod
    def julian_day(dt):
        """Calculate the Julian Day from a datetime.datetime object in UTC."""

        # year and month numbers
        yr, mo, dy, hr, mn, sc, us = _sp.calendar_time(dt)
        if mo <= 2:  # From paper: "if M = 1 or 2, then Y = Y - 1 and M = M + 12"
            mo += 12
            yr -= 1
        # day of the month with decimal time
        dy = dy + hr/24.0 + mn/(24.0*60.0) + sc / \
            (24.0*60.0*60.0) + us/(24.0*60.0*60.0*1e6)
        # b is equal to 0 for the julian calendar and is equal to (2- A +
        # INT(A/4)), A = INT(Y/100), for the gregorian calendar
        a = int(yr / 100)
        b = 2 - a + int(a / 4)
        jd = int(365.25 * (yr + 4716)) + \
            int(30.6001 * (mo + 1)) + dy + b - 1524.5
        return jd

    @staticmethod
    def julian_ephemeris_day(jd, deltat):
        """Calculate the Julian Ephemeris Day from the Julian Day and delta-time = (terrestrial time - universal time) in seconds."""

        return jd + deltat / 86400.0

    @staticmethod
    def julian_century(jd):
        """Caluclate the Julian Century from Julian Day or Julian Ephemeris Day."""

        return (jd - 2451545.0) / 36525.0

    @staticmethod
    def julian_millennium(jc):
        """Calculate the Julian Millennium from Julian Ephemeris Century."""

        return jc / 10.0

    # Earth Periodic Terms
    # Earth Heliocentric Longitude coefficients (L0, L1, L2, L3, L4, and L5 in paper)
    _EHL_ = [  # L0:
        [(175347046, 0.0, 0.0), (3341656, 4.6692568, 6283.07585), (34894, 4.6261, 12566.1517),
         (3497, 2.7441, 5753.3849), (3418, 2.8289,
                                     3.5231), (3136, 3.6277, 77713.7715),
         (2676, 4.4181, 7860.4194), (2343, 6.1352,
                                     3930.2097), (1324, 0.7425, 11506.7698),
         (1273, 2.0371, 529.691), (1199, 1.1096,
                                   1577.3435), (990, 5.233, 5884.927),
         (902, 2.045, 26.298), (857, 3.508, 398.149), (780, 1.179, 5223.694),
         (753, 2.533, 5507.553), (505, 4.583, 18849.228), (492, 4.205, 775.523),
         (357, 2.92, 0.067), (317, 5.849, 11790.629), (284, 1.899, 796.298),
         (271, 0.315, 10977.079), (243, 0.345, 5486.778), (206, 4.806, 2544.314),
         (205, 1.869, 5573.143), (202, 2.4458, 6069.777), (156, 0.833, 213.299),
         (132, 3.411, 2942.463), (126, 1.083, 20.775), (115, 0.645, 0.98),
         (103, 0.636, 4694.003), (102, 0.976, 15720.839), (102, 4.267, 7.114),
         (99, 6.21, 2146.17), (98, 0.68, 155.42), (86, 5.98, 161000.69),
         (85, 1.3, 6275.96), (85, 3.67, 71430.7), (80, 1.81, 17260.15),
         (79, 3.04, 12036.46), (71, 1.76, 5088.63), (74, 3.5, 3154.69),
         (74, 4.68, 801.82), (70, 0.83, 9437.76), (62, 3.98, 8827.39),
         (61, 1.82, 7084.9), (57, 2.78, 6286.6), (56, 4.39, 14143.5),
         (56, 3.47, 6279.55), (52, 0.19, 12139.55), (52, 1.33, 1748.02),
         (51, 0.28, 5856.48), (49, 0.49, 1194.45), (41, 5.37, 8429.24),
         (41, 2.4, 19651.05), (39, 6.17, 10447.39), (37, 6.04, 10213.29),
         (37, 2.57, 1059.38), (36, 1.71, 2352.87), (36, 1.78, 6812.77),
         (33, 0.59, 17789.85), (30, 0.44, 83996.85), (30, 2.74, 1349.87),
         (25, 3.16, 4690.48)],
        # L1:
        [(628331966747, 0.0, 0.0), (206059, 2.678235, 6283.07585), (4303, 2.6351, 12566.1517),
         (425, 1.59, 3.523), (119, 5.796, 26.298), (109, 2.966, 1577.344),
         (93, 2.59, 18849.23), (72, 1.14, 529.69), (68, 1.87, 398.15),
         (67, 4.41, 5507.55), (59, 2.89, 5223.69), (56, 2.17, 155.42),
         (45, 0.4, 796.3), (36, 0.47, 775.52), (29, 2.65, 7.11),
         (21, 5.34, 0.98), (19, 1.85, 5486.78), (19, 4.97, 213.3),
         (17, 2.99, 6275.96), (16, 0.03, 2544.31), (16, 1.43, 2146.17),
         (15, 1.21, 10977.08), (12, 2.83, 1748.02), (12, 3.26, 5088.63),
         (12, 5.27, 1194.45), (12, 2.08, 4694), (11, 0.77, 553.57),
         (10, 1.3, 3286.6), (10, 4.24, 1349.87), (9, 2.7, 242.73),
         (9, 5.64, 951.72), (8, 5.3, 2352.87), (6, 2.65, 9437.76),
         (6, 4.67, 4690.48)],
        # L2:
        [(52919, 0.0, 0.0), (8720, 1.0721, 6283.0758), (309, 0.867, 12566.152),
         (27, 0.05, 3.52), (16, 5.19, 26.3), (16, 3.68, 155.42),
         (10, 0.76, 18849.23), (9, 2.06, 77713.77), (7, 0.83, 775.52),
         (5, 4.66, 1577.34), (4, 1.03, 7.11), (4, 3.44, 5573.14),
         (3, 5.14, 796.3), (3, 6.05, 5507.55), (3, 1.19, 242.73),
         (3, 6.12, 529.69), (3, 0.31, 398.15), (3, 2.28, 553.57),
         (2, 4.38, 5223.69), (2, 3.75, 0.98)],
        # L3:
        [(289, 5.844, 6283.076), (35, 0.0, 0.0,), (17, 5.49, 12566.15),
         (3, 5.2, 155.42), (1, 4.72, 3.52), (1, 5.3, 18849.23),
         (1, 5.97, 242.73)],
        # L4:
        [(114, 3.142, 0.0), (8, 4.13, 6283.08), (1, 3.84, 12566.15)],
        # L5:
        [(1, 3.14, 0.0)]
    ]

    # Earth Heliocentric Longitude coefficients (B0 and B1 in paper)
    _EHB_ = [  # B0:
        [(280, 3.199, 84334.662), (102, 5.422, 5507.553), (80, 3.88, 5223.69),
         (44, 3.7, 2352.87), (32, 4.0, 1577.34)],
        # B1:
        [(9, 3.9, 5507.55), (6, 1.73, 5223.69)]
    ]

    # Earth Heliocentric Radius coefficients (R0, R1, R2, R3, R4)
    _EHR_ = [  # R0:
        [(100013989, 0.0, 0.0), (1670700, 3.0984635, 6283.07585), (13956, 3.05525, 12566.1517),
         (3084, 5.1985, 77713.7715), (1628, 1.1739,
                                      5753.3849), (1576, 2.8469, 7860.4194),
         (925, 5.453, 11506.77), (542, 4.564, 3930.21), (472, 3.661, 5884.927),
         (346, 0.964, 5507.553), (329, 5.9, 5223.694), (307, 0.299, 5573.143),
         (243, 4.273, 11790.629), (212, 5.847, 1577.344), (186, 5.022, 10977.079),
         (175, 3.012, 18849.228), (110, 5.055, 5486.778), (98, 0.89, 6069.78),
         (86, 5.69, 15720.84), (86, 1.27, 161000.69), (85, 0.27, 17260.15),
         (63, 0.92, 529.69), (57, 2.01, 83996.85), (56, 5.24, 71430.7),
         (49, 3.25, 2544.31), (47, 2.58, 775.52), (45, 5.54, 9437.76),
         (43, 6.01, 6275.96), (39, 5.36, 4694), (38, 2.39, 8827.39),
         (37, 0.83, 19651.05), (37, 4.9, 12139.55), (36, 1.67, 12036.46),
         (35, 1.84, 2942.46), (33, 0.24, 7084.9), (32, 0.18, 5088.63),
         (32, 1.78, 398.15), (28, 1.21, 6286.6), (28, 1.9, 6279.55),
         (26, 4.59, 10447.39)],
        # R1:
        [(103019, 1.10749, 6283.07585), (1721, 1.0644, 12566.1517), (702, 3.142, 0.0),
         (32, 1.02, 18849.23), (31, 2.84, 5507.55), (25, 1.32, 5223.69),
         (18, 1.42, 1577.34), (10, 5.91, 10977.08), (9, 1.42, 6275.96),
         (9, 0.27, 5486.78)],
        # R2:
        [(4359, 5.7846, 6283.0758), (124, 5.579, 12566.152), (12, 3.14, 0.0),
         (9, 3.63, 77713.77), (6, 1.87, 5573.14), (3, 5.47, 18849)],
        # R3:
        [(145, 4.273, 6283.076), (7, 3.92, 12566.15)],
        # R4:
        [(4, 2.56, 6283.08)]
    ]

    @staticmethod
    def heliocentric_longitude(jme):
        """Compute the Earth Heliocentric Longitude (L) in degrees given the Julian Ephemeris Millennium."""

        #L5, ..., L0
        Li = [sum(a*np.cos(b + c*jme) for a, b, c in abcs)
              for abcs in reversed(_sp._EHL_)]
        L = np.polyval(Li, jme) / 1e8
        L = np.rad2deg(L) % 360
        return L

    @staticmethod
    def heliocentric_latitude(jme):
        """Compute the Earth Heliocentric Latitude (B) in degrees given the Julian Ephemeris Millennium."""

        Bi = [sum(a*np.cos(b + c*jme) for a, b, c in abcs)
              for abcs in reversed(_sp._EHB_)]
        B = np.polyval(Bi, jme) / 1e8
        B = np.rad2deg(B) % 360
        return B

    @staticmethod
    def heliocentric_radius(jme):
        """Compute the Earth Heliocentric Radius (R) in astronimical units given the Julian Ephemeris Millennium."""

        Ri = [sum(a*np.cos(b + c*jme) for a, b, c in abcs)
              for abcs in reversed(_sp._EHR_)]
        R = np.polyval(Ri, jme) / 1e8
        return R

    @staticmethod
    def heliocentric_position(jme):
        """Compute the Earth Heliocentric Longitude, Latitude, and Radius given the Julian Ephemeris Millennium.

        Returns (L, B, R) where L = longitude in degrees, B = latitude in degrees, and R = radius in astronimical units."""

        return _sp.heliocentric_longitude(jme), _sp.heliocentric_latitude(jme), _sp.heliocentric_radius(jme)

    @staticmethod
    def geocentric_position(helio_pos):
        """Compute the geocentric latitude (Theta) and longitude (beta) (in degrees) of the sun given Earth's heliocentric position (L, B, R)."""

        L, B, R = helio_pos
        th = L + 180
        b = -B
        return (th, b)

    # Nutation Longitude and Obliquity coefficients (Y)
    _NLOY_ = [(0,   0,   0,   0,   1), (-2,  0,   0,   2,   2), (0,   0,   0,   2,   2),
              (0,   0,   0,   0,   2), (0,   1,   0,
                                        0,   0), (0,   0,   1,   0,   0),
              (-2,  1,   0,   2,   2), (0,   0,   0,
                                        2,   1), (0,   0,   1,   2,   2),
              (-2,  -1,  0,   2,   2), (-2,  0,   1,
                                        0,   0), (-2,  0,   0,   2,   1),
              (0,   0,   -1,  2,   2), (2,   0,   0,
                                        0,   0), (0,   0,   1,   0,   1),
              (2,   0,   -1,  2,   2), (0,   0,   -
                                        1,  0,   1), (0,   0,   1,   2,   1),
              (-2,  0,   2,   0,   0), (0,   0,   -
                                        2,  2,   1), (2,   0,   0,   2,   2),
              (0,   0,   2,   2,   2), (0,   0,   2,
                                        0,   0), (-2,  0,   1,   2,   2),
              (0,   0,   0,   2,   0), (-2,  0,   0,
                                        2,   0), (0,   0,   -1,  2,   1),
              (0,   2,   0,   0,   0), (2,   0,   -
                                        1,  0,   1), (-2,  2,   0,   2,   2),
              (0,   1,   0,   0,   1), (-2,  0,   1,
                                        0,   1), (0,   -1,  0,   0,   1),
              (0,   0,   2,   -2,  0), (2,   0,   -
                                        1,  2,   1), (2,   0,   1,   2,   2),
              (0,   1,   0,   2,   2), (-2,  1,   1,
                                        0,   0), (0,   -1,  0,   2,   2),
              (2,   0,   0,   2,   1), (2,   0,   1,
                                        0,   0), (-2,  0,   2,   2,   2),
              (-2,  0,   1,   2,   1), (2,   0,   -
                                        2,  0,   1), (2,   0,   0,   0,   1),
              (0,   -1,  1,   0,   0), (-2,  -1,  0,
                                        2,   1), (-2,  0,   0,   0,   1),
              (0,   0,   2,   2,   1), (-2,  0,   2,
                                        0,   1), (-2,  1,   0,   2,   1),
              (0,   0,   1,   -2,  0), (-1,  0,   1,
                                        0,   0), (-2,  1,   0,   0,   0),
              (1,   0,   0,   0,   0), (0,   0,   1,
                                        2,   0), (0,   0,   -2,  2,   2),
              (-1,  -1,  1,   0,   0), (0,   1,   1,
                                        0,   0), (0,   -1,  1,   2,   2),
              (2,   -1,  -1,  2,   2), (0,   0,   3,   2,   2), (2,   -1,  0,   2,   2)]
    # Nutation Longitude and Obliquity coefficients (a,b)
    _NLOab_ = [(-171996, -174.2), (-13187, -1.6), (-2274, -0.2), (2062, 0.2), (1426, -3.4), (712, 0.1),
               (-517, 1.2), (-386, -0.4), (-301,
                                           0), (217, -0.5), (-158, 0), (129, 0.1),
               (123, 0), (63,  0), (63,  0.1), (-59, 0), (-58, -0.1), (-51, 0),
               (48,  0), (46,  0), (-38, 0), (-31, 0), (29,  0), (29,  0),
               (26,  0), (-22, 0), (21,  0), (17,  -0.1), (16,  0), (-16, 0.1),
               (-15, 0), (-13, 0), (-12, 0), (11,  0), (-10, 0), (-8,  0),
               (7,   0), (-7,  0), (-7,  0), (-7,  0), (6,   0), (6,   0),
               (6,   0), (-6,  0), (-6,  0), (5,   0), (-5,  0), (-5,  0),
               (-5,  0), (4,   0), (4,   0), (4,   0), (-4,  0), (-4,  0),
               (-4,  0), (3,   0), (-3,  0), (-3,  0), (-3,  0), (-3,  0),
               (-3,  0), (-3,  0), (-3,  0)]
    # Nutation Longitude and Obliquity coefficients (c,d)
    _NLOcd_ = [(92025,   8.9), (5736,    -3.1), (977, -0.5), (-895,    0.5),
               (54,  -0.1), (-7,  0), (224, -0.6), (200, 0),
               (129, -0.1), (-95, 0.3), (0,   0), (-70, 0),
               (-53, 0), (0,   0), (-33, 0), (26,  0),
               (32,  0), (27,  0), (0,   0), (-24, 0),
               (16,  0), (13,  0), (0,   0), (-12, 0),
               (0,   0), (0,   0), (-10, 0), (0,   0),
               (-8,  0), (7,   0), (9,   0), (7,   0),
               (6,   0), (0,   0), (5,   0), (3,   0),
               (-3,  0), (0,   0), (3,   0), (3,   0),
               (0,   0), (-3,  0), (-3,  0), (3,   0),
               (3,   0), (0,   0), (3,   0), (3,   0),
               (3,   0)]

    @staticmethod
    def ecliptic_obliquity(jme, delta_epsilon):
        """Calculate the true obliquity of the ecliptic (epsilon, in degrees) given the Julian Ephemeris Millennium and the obliquity."""

        u = jme/10
        e0 = np.polyval([2.45, 5.79, 27.87, 7.12, -39.05, -
                         249.67, -51.38, 1999.25, -1.55, -4680.93, 84381.448], u)
        e = e0/3600.0 + delta_epsilon
        return e

    @staticmethod
    def nutation_obliquity(jce):
        """Compute the nutation in longitude (delta_psi) and the true obliquity (epsilon) given the Julian Ephemeris Century."""

        # mean elongation of the moon from the sun, in radians:
        #x0 = 297.85036 + 445267.111480*jce - 0.0019142*(jce**2) + (jce**3)/189474
        x0 = np.deg2rad(np.polyval(
            [1./189474, -0.0019142, 445267.111480, 297.85036], jce))
        # mean anomaly of the sun (Earth), in radians:
        x1 = np.deg2rad(np.polyval(
            [-1/3e5, -0.0001603, 35999.050340, 357.52772], jce))
        # mean anomaly of the moon, in radians:
        x2 = np.deg2rad(np.polyval(
            [1./56250, 0.0086972, 477198.867398, 134.96298], jce))
        # moon's argument of latitude, in radians:
        x3 = np.deg2rad(np.polyval(
            [1./327270, -0.0036825, 483202.017538, 93.27191], jce))
        # Longitude of the ascending node of the moon's mean orbit on the ecliptic
        # measured from the mean equinox of the date, in radians
        x4 = np.deg2rad(np.polyval(
            [1./45e4, 0.0020708, -1934.136261, 125.04452], jce))

        x = (x0, x1, x2, x3, x4)

        dp = 0.0
        for y, ab in zip(_sp._NLOY_, _sp._NLOab_):
            a, b = ab
            dp += (a + b*jce)*np.sin(np.dot(x, y))
        dp = np.rad2deg(dp)/36e6

        de = 0.0
        for y, cd in zip(_sp._NLOY_, _sp._NLOcd_):
            c, d = cd
            de += (c + d*jce)*np.cos(np.dot(x, y))
        de = np.rad2deg(de)/36e6

        e = _sp.ecliptic_obliquity(_sp.julian_millennium(jce), de)

        return dp, e

    @staticmethod
    def abberation_correction(R):
        """Calculate the abberation correction (delta_tau, in degrees) given the Earth Heliocentric Radius (in AU)."""

        return -20.4898/(3600*R)

    @staticmethod
    def sun_longitude(helio_pos, delta_psi):
        """Calculate the apparent sun longitude (lambda, in degrees) and geocentric longitude (beta, in degrees) given the earth heliocentric position and delta_psi."""

        L, B, R = helio_pos
        theta = L + 180  # geocentric latitude
        beta = -B
        ll = theta + delta_psi + _sp.abberation_correction(R)
        return ll, beta

    @staticmethod
    def greenwich_sidereal_time(jd, delta_psi, epsilon):
        """Calculate the apparent Greenwich sidereal time (v, in degrees) given the Julian Day."""

        jc = _sp.julian_century(jd)
        # mean sidereal time at greenwich, in degrees:
        v0 = (280.46061837 + 360.98564736629*(jd - 2451545) +
              0.000387933*(jc**2) - (jc**3)/38710000) % 360
        v = v0 + delta_psi*np.cos(np.deg2rad(epsilon))
        return v

    @staticmethod
    def sun_ra_decl(llambda, epsilon, beta):
        """Calculate the sun's geocentric right ascension (alpha, in degrees) and declination (delta, in degrees)."""

        l, e, b = map(np.deg2rad, (llambda, epsilon, beta))
        alpha = np.arctan2(np.sin(l)*np.cos(e) - np.tan(b)
                           * np.sin(e), np.cos(l))  # x1 / x2
        alpha = np.rad2deg(alpha) % 360
        delta = np.arcsin(np.sin(b)*np.cos(e) + np.cos(b)*np.sin(e)*np.sin(l))
        delta = np.rad2deg(delta)
        return alpha, delta

    @staticmethod
    def sun_topo_ra_decl_hour(latitude, longitude, elevation, jd, delta_t=0):
        """Calculate the sun's topocentric right ascension (alpha'), declination (delta'), and hour angle (H')."""

        jde = _sp.julian_ephemeris_day(jd, delta_t)
        jce = _sp.julian_century(jde)
        jme = _sp.julian_millennium(jce)

        helio_pos = _sp.heliocentric_position(jme)
        R = helio_pos[-1]
        phi, sigma, E = latitude, longitude, elevation
        # equatorial horizontal parallax of the sun, in radians
        xi = np.deg2rad(8.794/(3600*R))
        # rho = distance from center of earth in units of the equatorial radius
        # phi-prime = geocentric latitude
        # NB: These equations look like their based on WGS-84, but are rounded slightly
        # The WGS-84 reference ellipsoid has major axis a = 6378137 m, and flattening factor 1/f = 298.257223563
        # minor axis b = a*(1-f) = 6356752.3142 = 0.996647189335*a
        u = np.arctan(0.99664719*np.tan(phi))
        x = np.cos(u) + E*np.cos(phi)/6378140  # rho sin(phi-prime)
        y = 0.99664719*np.sin(u) + E*np.sin(phi)/6378140  # rho cos(phi-prime)

        delta_psi, epsilon = _sp.nutation_obliquity(jce)

        llambda, beta = _sp.sun_longitude(helio_pos, delta_psi)

        alpha, delta = _sp.sun_ra_decl(llambda, epsilon, beta)

        v = _sp.greenwich_sidereal_time(jd, delta_psi, epsilon)

        H = v + longitude - alpha
        Hr, dr = map(np.deg2rad, (H, delta))

        dar = np.arctan2(-x*np.sin(xi)*np.sin(Hr),
                         np.cos(dr)-x*np.sin(xi)*np.cos(Hr))
        delta_alpha = np.rad2deg(dar)

        alpha_prime = alpha + delta_alpha
        delta_prime = np.rad2deg(np.arctan2(
            (np.sin(dr) - y*np.sin(xi))*np.cos(dar), np.cos(dr) - y*np.sin(xi)*np.cos(Hr)))
        H_prime = H - delta_alpha

        return alpha_prime, delta_prime, H_prime

    @staticmethod
    def sun_topo_azimuth_zenith(latitude, delta_prime, H_prime, temperature=14.6, pressure=1013):
        """Compute the sun's topocentric azimuth and zenith angles.

        Azimuth is measured eastward from north, zenith from vertical.
        Temperature = average temperature in C (default is 14.6 = global average in 2013).
        Pressure = average pressure in mBar (default 1013 = global average).
        """

        phi = np.deg2rad(latitude)
        dr, Hr = map(np.deg2rad, (delta_prime, H_prime))
        P, T = pressure, temperature
        e0 = np.rad2deg(np.arcsin(np.sin(phi)*np.sin(dr) +
                                  np.cos(phi)*np.cos(dr)*np.cos(Hr)))
        tmp = np.deg2rad(e0 + 10.3/(e0+5.11))
        delta_e = (P/1010.0)*(283.0/(273+T))*(1.02/(60*np.tan(tmp)))
        e = e0 + delta_e
        zenith = 90 - e

        gamma = np.rad2deg(np.arctan2(np.sin(Hr), np.cos(
            Hr)*np.sin(phi) - np.tan(dr)*np.cos(phi))) % 360
        Phi = (gamma + 180) % 360  # azimuth from north
        return Phi, zenith

    @staticmethod
    def norm_lat_lon(lat, lon):
        """."""

        if lat < -90 or lat > 90:
            # convert to cartesian and back
            x = np.cos(np.deg2rad(lon))*np.cos(np.deg2rad(lat))
            y = np.sin(np.deg2rad(lon))*np.cos(np.deg2rad(lat))
            z = np.sin(np.deg2rad(lat))
            r = np.sqrt(x**2 + y**2 + z**2)
            lon = np.rad2deg(np.arctan2(y, x)) % 360
            lat = np.rad2deg(np.arcsin(z/r))
        elif lon < 0 or lon > 360:
            lon = lon % 360
        return lat, lon

    @staticmethod
    def topo_pos(t, lat, lon, elev, temp, press, dt):
        """Compute RA,dec,H, all in degrees."""

        lat, lon = _sp.norm_lat_lon(lat, lon)
        jd = _sp.julian_day(t)
        RA, dec, H = _sp.sun_topo_ra_decl_hour(lat, lon, elev, jd, dt)
        return RA, dec, H

    @staticmethod
    def pos(t, lat, lon, elev, temp, press, dt):
        """Compute azimute,zenith,RA,dec,H all in degree."""

        lat, lon = _sp.norm_lat_lon(lat, lon)
        jd = _sp.julian_day(t)
        RA, dec, H = _sp.sun_topo_ra_decl_hour(lat, lon, elev, jd, dt)
        azimuth, zenith = _sp.sun_topo_azimuth_zenith(lat, dec, H, temp, press)
        return azimuth, zenith, RA, dec, H


def julian_day(dt):
    """Convert UTC datetimes or UTC timestamps to Julian days.

    Parameters
    ----------
    dt : array_like
        UTC datetime objects or UTC timestamps (as per datetime.utcfromtimestamp)

    Returns
    -------
    jd : ndarray
        datetimes converted to fractional Julian days
    """

    dts = np.array(dt)
    if len(dts.shape) == 0:
        return _sp.julian_day(dt)

    jds = np.empty(dts.shape)
    for i, d in enumerate(dts.flat):
        jds.flat[i] = _sp.julian_day(d)
    return jds


def arcdist(p0, p1, radians=False):
    """Angular distance between azimuth, zenith pairs.

    Parameters
    ----------
    p0 : array_like, shape (..., 2)
    p1 : array_like, shape (..., 2)
        p[...,0] = azimuth angles, p[...,1] = zenith angles
    radians : boolean (default False)
        If False, angles are in degrees, otherwise in radians

    Returns
    -------
    ad :  array_like, shape is broadcast(p0,p1).shape
        Arcdistances between corresponding pairs in p0,p1
        In degrees by default, in radians if radians=True
    """

    # formula comes from translating points into cartesian coordinates
    # taking the dot product to get the cosine between the two vectors
    # then arccos to return to angle, and simplify everything assuming real inputs
    p0, p1 = np.array(p0), np.array(p1)
    if not radians:
        p0, p1 = np.deg2rad(p0), np.deg2rad(p1)
    a0, z0 = p0[..., 0], p0[..., 1]
    a1, z1 = p1[..., 0], p1[..., 1]
    d = np.arccos(np.cos(z0)*np.cos(z1)+np.cos(a0-a1)*np.sin(z0)*np.sin(z1))
    if radians:
        return d
    else:
        return np.rad2deg(d)


def observed_sunpos(dt, latitude, longitude, elevation, temperature=None, pressure=None, delta_t=0, radians=False):
    """Compute the observed coordinates of the sun as viewed at the given time and location.

    Parameters
    ----------
    dt : array_like
        UTC datetime objects or UTC timestamps (as per datetime.utcfromtimestamp) representing the times of observations
    latitude, longitude : array_like
        decimal degrees, positive for north of the equator and east of Greenwich
    elevation : array_like
        meters, relative to the WGS-84 ellipsoid
    temperature : array_like or None, optional
        celcius, default is 14.6 (global average in 2013)
    pressure : array_like or None, optional
        millibar, default is 1013 (global average in ??)
    delta_t : array_like, optional
        seconds, default is 0, difference between the earth's rotation time (TT) and universal time (UT)
    radians : {True, False}, optional
        return results in radians if True, degrees if False (default)

    Returns
    -------
    coords : ndarray, (...,2)
        The shape of the array is parameters broadcast together, plus a final dimension for the coordinates.
        coords[...,0] = observed azimuth angle, measured eastward from north
        coords[...,1] = observed zenith angle, measured down from vertical
    """

    if temperature is None:
        temperature = 14.6
    if pressure is None:
        pressure = 1013

    # 6367444 = radius of earth
    # numpy broadcasting
    b = np.broadcast(dt, latitude, longitude, elevation,
                     temperature, pressure, delta_t)
    res = np.empty(b.shape+(2,))
    res_vec = res.reshape((-1, 2))
    for i, x in enumerate(b):
        res_vec[i] = _sp.pos(*x)[:2]
    if radians:
        res = np.deg2rad(res)
    return res


def topocentric_sunpos(dt, latitude, longitude, temperature=None, pressure=None, delta_t=0, radians=False):
    """Compute the topocentric coordinates of the sun as viewed at the given time and location.

    Parameters
    ----------
    dt : array_like
        UTC datetime objects or UTC timestamps (as per datetime.utcfromtimestamp) representing the times of observations
    latitude, longitude : array_like
        decimal degrees, positive for north of the equator and east of Greenwich
    elevation : array_like
        meters, relative to the WGS-84 ellipsoid
    temperature : array_like or None, optional
        celcius, default is 14.6 (global average in 2013)
    pressure : array_like or None, optional
        millibar, default is 1013 (global average in ??)
    delta_t : array_like, optional
        seconds, default is 0, difference between the earth's rotation time (TT) and universal time (UT)
    radians : {True, False}, optional
        return results in radians if True, degrees if False (default)

    Returns
    -------
    coords : ndarray, (...,3)
        The shape of the array is parameters broadcast together, plus a final dimension for the coordinates.
        coords[...,0] = topocentric right ascension
        coords[...,1] = topocentric declination
        coords[...,2] = topocentric hour angle
    """

    if temperature is None:
        temperature = 14.6
    if pressure is None:
        pressure = 1013

    # 6367444 = radius of earth
    # numpy broadcasting
    b = np.broadcast(dt, latitude, longitude, elevation,
                     temperature, pressure, delta_t)
    res = np.empty(b.shape+(2,))
    res_vec = res.reshape((-1, 2))
    for i, x in enumerate(b):
        res_vec[i] = _sp.topo_pos(*x)
    if radians:
        res = np.deg2rad(res)
    return res


def sunpos(dt, latitude, longitude, elevation, temperature=None, pressure=None, delta_t=0, radians=False):
    """Compute the observed and topocentric coordinates of the sun as viewed at the given time and location.

    Parameters
    ----------
    dt : array_like
        UTC datetime objects or UTC timestamps (as per datetime.utcfromtimestamp) representing the times of observations
    latitude, longitude : array_like
        decimal degrees, positive for north of the equator and east of Greenwich
    elevation : array_like
        meters, relative to the WGS-84 ellipsoid
    temperature : array_like or None, optional
        celcius, default is 14.6 (global average in 2013)
    pressure : array_like or None, optional
        millibar, default is 1013 (global average in ??)
    delta_t : array_like, optional
        seconds, default is 0, difference between the earth's rotation time (TT) and universal time (UT)
    radians : {True, False}, optional
        return results in radians if True, degrees if False (default)

    Returns
    -------
    coords : ndarray, (...,5)
        The shape of the array is parameters broadcast together, plus a final dimension for the coordinates.
        coords[...,0] = observed azimuth angle, measured eastward from north
        coords[...,1] = observed zenith angle, measured down from vertical
        coords[...,2] = topocentric right ascension
        coords[...,3] = topocentric declination
        coords[...,4] = topocentric hour angle
    """

    if temperature is None:
        temperature = 14.6
    if pressure is None:
        pressure = 1013

    # 6367444 = radius of earth
    # numpy broadcasting
    b = np.broadcast(dt, latitude, longitude, elevation,
                     temperature, pressure, delta_t)
    res = np.empty(b.shape+(5,))
    res_vec = res.reshape((-1, 5))
    for i, x in enumerate(b):
        res_vec[i] = _sp.pos(*x)
    if radians:
        res = np.deg2rad(res)
    return res


class Sunposition:
    """Compute sun position parameters given the time and location."""

    # Inputs
    t = None
    lat, lon = None, None
    elev = None
    temp = None
    p = None
    dt = None
    rad = None
    # Outputs
    az = None
    zen = None
    ra = None
    dec = None
    h = None

    def __init__(self, t, lat, lon, elev, temp, p, dt, rad, csv=False):
        """Initialize the class and run the model."""

        self.lat = lat
        self.lon = lon
        self.elev = elev
        self.temp = temp
        self.p = p
        self.dt = dt
        self.rad = rad

        if t == "now":
            self.t = datetime.utcnow()
        elif ":" in t and "-" in t:
            try:
                # with microseconds
                self.t = datetime.strptime(t, '%Y-%m-%d %H:%M:%S.%f')
            except BaseException:
                try:
                    # without microseconds
                    self.t = datetime.strptime(t, '%Y-%m-%d %H:%M:%S.')
                except BaseException:
                    self.t = datetime.strptime(t, '%Y-%m-%d %H:%M:%S')
        else:
            self.t = datetime.utcfromtimestamp(int(t))

        # Run the sun position calculation
        self.az, self.zen, self.ra, self.dec, self.h = sunpos(
            self.t, lat, lon, elev, temp, p, dt, rad)

        # Format output to CSV?
        if csv:
            print('{t}, {dt}, {lat}, {lon}, {elev}, {temp}, {p}, {az}, {zen}, {ra}, {dec}, {h}'.format(
                t=self.t, dt=dt, lat=lat, lon=lon, elev=elev, temp=temp, p=p, az=self.az, zen=self.zen, ra=self.ra, dec=self.dec, h=self.h))
        else:
            dr = 'deg'
            if rad:
                dr = 'rad'
            print(
                "Computing sun position at T = {t} + {dt} s".format(t=self.t, dt=dt))
            print("Lat, Lon, Elev = {lat} deg, {lon} deg, {elev} m".format(
                lat=lat, lon=lon, elev=elev))
            print("T, P = {temp} C, {press} mbar".format(temp=temp, press=p))
            print("Results:")
            print("Azimuth, zenith = {az} {dr}, {zen} {dr}".format(
                az=self.az, zen=self.zen, dr=dr))
            print("RA, dec, H = {ra} {dr}, {dec} {dr}, {h} {dr}".format(
                ra=self.ra, dec=self.dec, h=self.h, dr=dr))

    @property
    def citation(self):
        """Print the citation."""

        print("Implementation: Samuel Bear Powell, 2016")
        print("Algorithm:")
        print("Ibrahim Reda, Afshin Andreas, \"Solar position algorithm for solar radiation applications\", SolarEnergy, Volume 76, Issue 5, 2004, Pages 577-589, ISSN 0038-092X, doi:10.1016/j.solener.2003.12.003")



#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import os
import logging
from datetime import datetime
import numpy as np

from isofit.core.common import resample_spectrum, load_wavelen, VectorInterpolator
from .look_up_tables import TabularRT, FileExistsError
from isofit.core.geometry import Geometry
from isofit.configs import Config
from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig

eps = 1e-5  # used for finite difference derivative calculations

sixs_template = '''0 (User defined)
{solzen} {solaz} {viewzen} {viewaz} {month} {day}
8  (User defined H2O, O3)
{H2OSTR}, {O3}
{aermodel}
0
{AOT550}
{elev:.2f} (target level)
-{alt:.2f} (sensor level)
-{H2OSTR}, -{O3}
{AOT550}
-2 
{wlinf}
{wlsup}
0 Homogeneous surface
0 (no directional effects)
0
0
0
-1 No atm. corrections selected
'''


class SixSRT(TabularRT):
    """A model of photon transport including the atmosphere."""

    def __init__(self, engine_config: RadiativeTransferEngineConfig, full_config: Config,
                 build_lut=True, build_lut_only=False, wavelength_override=None, fwhm_override=None,
                 modtran_emulation=False):

        self.angular_lut_keys_degrees = ['OBSZEN', 'TRUEAZ', 'viewzen', 'viewaz',
                                         'solzen', 'solaz']
        self.angular_lut_keys_radians = []
        self.modtran_emulation = modtran_emulation

        super().__init__(engine_config, full_config)

        self.treat_as_emissive = False
        self.lut_quantities = ['rhoatm', 'transm', 'sphalb', 'transup']

        if wavelength_override is not None:
            self.wl = wavelength_override
            self.n_chan = len(self.wl)
            self.resample_wavelengths = False
        else:
            self.resample_wavelengths = True

        if fwhm_override is not None:
            self.fwhm = fwhm_override

        self.sixs_dir = self.find_basedir(engine_config)
        self.sixs_grid_init = np.arange(self.wl[0], self.wl[-1]+2.5, 2.5)
        self.sixs_ngrid_init = len(self.sixs_grid_init)
        self.params = {'aermodel': 1,
                       'AOT550': 0.01,
                       'H2OSTR': 0,
                       'O3': 0.30,
                       'day':   engine_config.day,
                       'month': engine_config.month,
                       'elev':  engine_config.elev,
                       'alt':   engine_config.alt,
                       'atm_file': None,
                       'abscf_data_directory': None,
                       'wlinf': self.sixs_grid_init[0]/1000.0,  # convert to nm
                       'wlsup': self.sixs_grid_init[-1]/1000.0}

        if engine_config.obs_file is not None:
            # A special case where we load the observation geometry
            # from a custom-crafted text file
            g = Geometry(obs=engine_config.obs_file)
            self.params['solzen'] = g.solar_zenith
            self.params['solaz'] = g.solar_azimuth
            self.params['viewzen'] = g.observer_zenith
            self.params['viewaz'] = g.observer_azimuth
        else:
            # We have to get geometry from somewhere, so we presume it is
            # in the configuration file.
            self.params['solzen'] = engine_config.solzen
            self.params['viewzen'] = engine_config.viewzen
            self.params['solaz'] = engine_config.solaz
            self.params['viewaz'] = engine_config.viewaz

        if build_lut_only is False:
            self.esd = np.loadtxt(engine_config.earth_sun_distance_file)
            dt = datetime(2000, self.params['month'], self.params['day'])
            self.day_of_year = dt.timetuple().tm_yday
            self.irr_factor = self.esd[self.day_of_year-1, 1]

            irr = np.loadtxt(engine_config.irradiance_file, comments='#')
            iwl, irr = irr.T
            irr = irr / 10.0  # convert, uW/nm/cm2
            irr = irr / self.irr_factor**2  # consider solar distance
            self.solar_irr = resample_spectrum(irr, iwl,  self.wl, self.fwhm)

        if build_lut:
            self.build_lut()

    def find_basedir(self, config: RadiativeTransferEngineConfig):
        """Seek out a sixs base directory."""

        if config.engine_base_dir is not None:
            return config.engine_base_dir

        try:
            return os.getenv('SIXS_DIR')
        except KeyError:
            logging.error('I could not find the SIXS base directory')
            raise KeyError('I could not find the SIXS base directory')


    def rebuild_cmd(self, point, fn):
        """."""

        # start with defaults
        vals = self.params.copy()
        for n, v in zip(self.lut_names, point):
            vals[n] = v

        # Translate a couple of special cases
        if 'H2OSTR' in self.lut_names:
            vals['h2o_mm'] = vals['H2OSTR']*10.0
        if 'GNDALT' in vals:
            vals['elev'] = vals['GNDALT']
        if 'H1ALT' in vals:
            vals['alt'] = vals['H1ALT']
        if 'TRUEAZ' in vals:
            vals['viewaz'] = vals['TRUEAZ']
        if 'OBSZEN' in vals:
            vals['viewzen'] = 180 - vals['OBSZEN']

        if self.modtran_emulation:
            if 'AERFRAC_2' in vals:
                vals['AOT550'] = vals['AERFRAC_2']

        if 'elev' in vals:
            vals['elev'] = vals['elev']*-1

        # Check rebuild conditions: LUT is missing or from a different config
        scriptfilename = 'LUT_'+fn+'.sh'
        scriptfilepath = os.path.join(self.lut_dir, scriptfilename)
        infilename = 'LUT_'+fn+'.inp'
        infilepath = os.path.join(self.lut_dir, infilename)
        outfilename = fn
        outfilepath = os.path.join(self.lut_dir, outfilename)
        if os.path.exists(outfilepath) and os.path.exists(infilepath):
            raise FileExistsError('Files exist')

        if self.sixs_dir is None:
            logging.error('Specify a SixS installation')
            raise KeyError('Specify a SixS installation')

        sixspath = self.sixs_dir+'/sixsV2.1'

        # write config files
        sixs_config_str = sixs_template.format(**vals)
        with open(infilepath, 'w') as f:
            f.write(sixs_config_str)

        # Write runscript file
        with open(scriptfilepath, 'w') as f:
            f.write('#!/usr/bin/bash\n')
            f.write('%s < %s > %s\n' % (sixspath, infilepath, outfilepath))
            f.write('cd $cwd\n')

        return 'bash '+scriptfilepath

    def load_rt(self, fn, resample=True):
        """Load the results of a SixS run."""

        with open(os.path.join(self.lut_dir, fn), 'r') as l:
            lines = l.readlines()

        with open(os.path.join(self.lut_dir, 'LUT_'+fn+'.inp'), 'r') as l:
            inlines = l.readlines()
            solzen = float(inlines[1].strip().split()[0])
        self.coszen = np.cos(solzen/360*2.0*np.pi)

        # Strip header
        for i, ln in enumerate(lines):
            if ln.startswith('*        trans  down   up'):
                lines = lines[(i + 1):(i + 1 + self.sixs_ngrid_init)]
                break

        solzens = np.zeros(len(lines))
        sphalbs = np.zeros(len(lines))
        transups = np.zeros(len(lines))
        transms = np.zeros(len(lines))
        rhoatms = np.zeros(len(lines))
        self.grid = np.zeros(len(lines))

        for i, ln in enumerate(lines):
            ln = ln.replace('*', ' ').strip()
            w, gt, scad, scau, salb, rhoa, swl, step, sbor, dsol, toar = \
                ln.split()

            self.grid[i] = float(w) * 1000.0  # convert to nm
            solzens[i] = float(solzen)
            sphalbs[i] = float(salb)
            transups[i] = 0.0  # float(scau)
            transms[i] = float(scau) * float(scad) * float(gt)
            rhoatms[i] = float(rhoa)

        if resample:
            solzens = resample_spectrum(solzens,  self.grid, self.wl, self.fwhm)
            rhoatms = resample_spectrum(rhoatms,  self.grid, self.wl, self.fwhm)
            transms = resample_spectrum(transms,  self.grid, self.wl, self.fwhm)
            sphalbs = resample_spectrum(sphalbs,  self.grid, self.wl, self.fwhm)
            transups = resample_spectrum(transups,  self.grid, self.wl, self.fwhm)

        results = {"solzen": solzens, "rhoatm": rhoatms, "transm": transms, "sphalb": sphalbs,
               "transup": transups}
        return results

    def ext550_to_vis(self, ext550):
        """."""

        return np.log(50.0) / (ext550 + 0.01159)

    def build_lut(self, rebuild=False):

        TabularRT.build_lut(self, rebuild)

        sixs_outputs = []
        for point, fn in zip(self.points, self.files):
            sixs_outputs.append(self.load_rt(fn,resample=self.resample_wavelengths))

        self.cache = {}
        dims_aug = self.lut_dims + [self.n_chan]
        for key in self.lut_quantities:
            temp = np.zeros(dims_aug, dtype=float)
            for sixs_output, point in zip(sixs_outputs, self.points):
                ind = [np.where(g == p)[0] for g, p in
                       zip(self.lut_grids, point)]
                ind = tuple(ind)
                temp[ind] = sixs_output[key]

            self.luts[key] = VectorInterpolator(self.lut_grids, temp,
                                                self.lut_interp_types)

    def _lookup_lut(self, point):
        ret = {}
        for key, lut in self.luts.items():
            ret[key] = np.array(lut(point)).ravel()
        return ret

    def get(self, x_RT: np.array, geom: Geometry):
        """ Get interpolated six-s results at a particular location

        Args:
            x_RT: radiative-transfer portion of the statevector
            geom: local geometry conditions for lookup

        Returns:
            interpolated modtran result

        """
        point = np.zeros((self.n_point,))
        for point_ind, name in enumerate(self.lut_grid_config):
            if name in self.statevector_names:
                ix = self.statevector_names.index(name)
                point[point_ind] = x_RT[ix]
            elif name == "elev":
                point[point_ind] = geom.surface_elevation_km
            elif name == "alt":
                point[point_ind] = geom.observer_altitude_km
            elif name == "viewzen":
                point[point_ind] = geom.observer_zenith
            elif name == "viewaz":
                point[point_ind] = geom.observer_azimuth
            elif name == "solaz":
                point[point_ind] = geom.solar_azimuth
            elif name == "solzen":
                point[point_ind] = geom.solar_zenith
            elif name == 'phi':
                point[point_ind] = geom.phi
            elif name == 'umu':
                point[point_ind] = geom.umu
            else:
                # If a variable is defined in the lookup table but not
                # specified elsewhere, we will default to the minimum
                point[point_ind] = min(self.lut_grid_config[name])

        return self._lookup_lut(point)

    def get_L_atm(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rho = r['rhoatm']
        rdn = rho / np.pi*(self.solar_irr * self.coszen)
        return rdn

    def get_L_down_transmitted(self, x_RT, geom):
        r = self.get(x_RT, geom)
        rdn = (self.solar_irr * self.coszen) / np.pi * r['transm']
        return rdn



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import time
import logging
import numpy as np
from collections import OrderedDict
from scipy.optimize import least_squares
import scipy.linalg

from isofit.core.common import svd_inv, svd_inv_sqrt, eps, combos, conditional_gaussian
from .inverse_simple import invert_simple
from isofit.configs import Config
from isofit.core.forward import ForwardModel
from isofit.configs.sections.implementation_config import InversionConfig


### Variables ###

error_code = -1


### Classes ###

class Inversion:

    def __init__(self, full_config: Config, forward: ForwardModel):
        """Initialization specifies retrieval subwindows for calculating
        measurement cost distributions."""

        config: InversionConfig = full_config.implementation.inversion

        self.lasttime = time.time()
        self.fm = forward
        self.hashtable = OrderedDict()  # Hash table for caching inverse matrices
        self.max_table_size = 500
        self.state_indep_S_hat = False

        self.windows = config.windows  # Retrieval windows
        self.mode = full_config.implementation.mode
        self.state_indep_S_hat = config.cressie_map_confidence

        # We calculate the instrument channel indices associated with the
        # retrieval windows using the initial instrument calibration.  These
        # window indices never change throughout the life of the object.
        self.winidx = np.array((), dtype=int)  # indices of retrieval windows
        for lo, hi in self.windows:
            idx = np.where(np.logical_and(self.fm.instrument.wl_init > lo,
                                          self.fm.instrument.wl_init < hi))[0]
            self.winidx = np.concatenate((self.winidx, idx), axis=0)
        self.counts = 0
        self.inversions = 0

        self.integration_grid = OrderedDict(config.integration_grid)
        self.grid_as_starting_points = config.inversion_grid_as_preseed

        if self.grid_as_starting_points:
            # We're using the integration grid to preseed, not fix values.  So
            # Track the grid, but don't fix the integration grid points
            self.inds_fixed = []
            self.inds_preseed = np.array([self.fm.statevec.index(k) for k in
                               self.integration_grid.keys()])
            self.inds_free = np.array([i for i in np.arange(self.fm.nstate, dtype=int) if
                              not (i in self.inds_fixed)])

        else:
            # We're using the integration grid to fix values.  So
            # Get set up to fix the integration grid points
            self.inds_fixed = np.array([self.fm.statevec.index(k) for k in
                               self.integration_grid.keys()])
            self.inds_free = np.array([i for i in np.arange(self.fm.nstate, dtype=int) if
                              not (i in self.inds_fixed)])
            self.inds_preseed = []

        self.x_fixed = None

        # Set least squares params that come from the forward model
        self.least_squares_params = {
            'method': 'trf',
            'max_nfev': 20,
            'bounds': (self.fm.bounds[0][self.inds_free],
                       self.fm.bounds[1][self.inds_free]),
            'x_scale': self.fm.scale[self.inds_free],
        }

        # Update the rest from the config
        for key, item in config.least_squares_params.get_config_options_as_dict().items():
            self.least_squares_params[key] = item

    def full_statevector(self, x_free):
        x = np.zeros(self.fm.nstate)
        if self.x_fixed is not None:
            x[self.inds_fixed] = self.x_fixed
        x[self.inds_free] = x_free
        return x

    def calc_conditional_prior(self, x_free, geom):
        """Calculate prior distribution of radiance. This depends on the
        location in the state space. Return the inverse covariance and
        its square root (for non-quadratic error residual calculation)."""

        x = self.full_statevector(x_free)
        xa = self.fm.xa(x, geom)
        Sa = self.fm.Sa(x, geom)

        # If there aren't any fixed parameters, we just directly
        if self.x_fixed is None or self.grid_as_starting_points:
            Sa_inv, Sa_inv_sqrt = svd_inv_sqrt(Sa, hashtable=self.hashtable)
            return xa, Sa, Sa_inv, Sa_inv_sqrt

        else:
            # otherwise condition on fixed variables
            #TODO: could make the below calculation without the svd_inv (using full initial inversion),
            # which would be way cheaper
            xa_free, Sa_free = conditional_gaussian(xa, Sa, self.inds_free,
                                                    self.inds_fixed, self.x_fixed)
            Sa_free_inv, Sa_free_inv_sqrt = svd_inv_sqrt(Sa_free,
                                                         hashtable=self.hashtable)
            return xa_free, Sa_free, Sa_free_inv, Sa_free_inv_sqrt


    def calc_prior(self, x, geom):
        """Calculate prior distribution of radiance. This depends on the 
        location in the state space. Return the inverse covariance and 
        its square root (for non-quadratic error residual calculation)."""

        xa = self.fm.xa(x, geom)
        Sa = self.fm.Sa(x, geom)
        Sa_inv, Sa_inv_sqrt = svd_inv_sqrt(Sa, hashtable=self.hashtable)
        return xa, Sa, Sa_inv, Sa_inv_sqrt

    def calc_posterior(self, x, geom, meas):
        """Calculate posterior distribution of state vector. This depends 
        both on the location in the state space and the radiance (via noise)."""

        xa = self.fm.xa(x, geom)
        Sa = self.fm.Sa(x, geom)
        Sa_inv = svd_inv(Sa, hashtable=self.hashtable)
        K = self.fm.K(x, geom)
        Seps = self.fm.Seps(x, meas, geom)
        Seps_inv = svd_inv(Seps, hashtable=self.hashtable)

        # Gain matrix G reflects current state, so we use the state-dependent
        # Jacobian matrix K
        S_hat = svd_inv(K.T.dot(Seps_inv).dot(
            K) + Sa_inv, hashtable=self.hashtable)
        G = S_hat.dot(K.T).dot(Seps_inv)

        # N. Cressie [ASA 2018] suggests an alternate definition of S_hat for
        # more statistically-consistent posterior confidence estimation
        if self.state_indep_S_hat:
            Ka = self.fm.K(xa, geom)
            S_hat = svd_inv(Ka.T.dot(Seps_inv).dot(Ka) + Sa_inv,
                            hashtable=self.hashtable)
        return S_hat, K, G

    def calc_Seps(self, x, meas, geom):
        """Calculate (zero-mean) measurement distribution in radiance terms.
        This depends on the location in the state space. This distribution is 
        calculated over one or more subwindows of the spectrum. Return the 
        inverse covariance and its square root."""

        Seps = self.fm.Seps(x, meas, geom)
        wn = len(self.winidx)
        Seps_win = np.zeros((wn, wn))
        for i in range(wn):
            Seps_win[i, :] = Seps[self.winidx[i], self.winidx]
        return svd_inv_sqrt(Seps_win, hashtable=self.hashtable)


    def jacobian(self, x_free, geom, Seps_inv_sqrt) -> np.ndarray:
        """Calculate measurement Jacobian and prior Jacobians with
        respect to cost function. This is the derivative of cost with
        respect to the state, commonly known as the gradient or loss
        surface. The cost is expressed as a vector of 'residuals'
        with respect to the prior and measurement, expressed in absolute
        (not quadratic) terms for the solver; It is the square root of
        the Rodgers (2000) Chi-square version. All measurement
        distributions are calculated over subwindows of the full
        spectrum.
        Args:
            x_free: decision variables - portion of the statevector not fixed by a static integration grid
            geom: Geometry to use for inversion
            Seps_inv_sqrt: Inverse square root of the covariance of "observation noise",
             including both measurement noise from the instrument as well as variability due to
             unknown variables.

        Returns:
            total_jac: The complete (measurement and prior) jacobian
        """
        x = self.full_statevector(x_free)

        # jacobian of measurment cost term WRT full state vector.
        K = self.fm.K(x, geom)[self.winidx, :]
        K = K[:, self.inds_free]
        meas_jac = Seps_inv_sqrt.dot(K)

        # jacobian of prior cost term with respect to state vector.
        xa_free, Sa_free, Sa_free_inv, Sa_free_inv_sqrt = \
            self.calc_conditional_prior(x_free, geom)
        prior_jac = Sa_free_inv_sqrt

        # The total cost vector (as presented to the solver) is the
        # concatenation of the "residuals" due to the measurement
        # and prior distributions. They will be squared internally by
        # the solver.
        total_jac = np.real(np.concatenate((meas_jac, prior_jac), axis=0))

        return total_jac

    def loss_function(self, x_free, geom, Seps_inv_sqrt, meas) -> (np.array, np.array):
        """Calculate cost function expressed here in absolute (not
        quadratic) terms for the solver, i.e., the square root of the
        Rodgers (2000) Chi-square version. We concatenate 'residuals'
        due to measurment and prior terms, suitably scaled.
        All measurement distributions are calculated over subwindows
        of the full spectrum.

        Args:
            x_free: decision variables - portion of the statevector not fixed by a static integration grid
            geom: Geometry to use for inversion
            Seps_inv_sqrt: Inverse square root of the covariance of "observation noise",
             including both measurement noise from the instrument as well as variability due to
             unknown variables.
            meas: a one-D scipy vector of radiance in uW/nm/sr/cm2

        Returns:
            total_residual: the complete, calculated residual
            x: the complete (x_free + any x_fixed augmented in)
        """

        # set up full-sized state vector
        x = self.full_statevector(x_free)

        # Measurement cost term.  Will calculate reflectance and Ls from
        # the state vector.
        est_meas = self.fm.calc_meas(x, geom, rfl=None, Ls=None)
        est_meas_window = est_meas[self.winidx]
        meas_window = meas[self.winidx]
        meas_resid = (est_meas_window - meas_window).dot(Seps_inv_sqrt)

        # Prior cost term
        xa_free, Sa_free, Sa_free_inv, Sa_free_inv_sqrt = \
            self.calc_conditional_prior(x_free, geom)
        prior_resid = (x_free - xa_free).dot(Sa_free_inv_sqrt)

        # Total cost
        total_resid = np.concatenate((meas_resid, prior_resid))

        return np.real(total_resid), x

    def invert(self, meas, geom):
        """Inverts a meaurement and returns a state vector.
        Args:
            meas: a one-D scipy vector of radiance in uW/nm/sr/cm2
            geom: a geometry object

        Returns:
            final_solution: a converged state vector solution
        """
        self.counts = 0
        costs, solutions = [], []

        # Simulations are easy - return the initial state vector
        if self.mode == 'simulation' or meas is None:
            return np.array([self.fm.init.copy()])

        if len(self.integration_grid.values()) == 0:
            combo_values = [None]
        else:
            combo_values = combos(self.integration_grid.values()).copy()

        for combo in combo_values:

            if self.grid_as_starting_points is False:
                self.x_fixed = combo
            trajectory = []

            # Calculate the initial solution, if needed.
            x0 = invert_simple(self.fm, meas, geom)
            x0 = x0[self.inds_free]

            # Catch any state vector elements outside of bounds
            lower_bound_violation = x0 < self.fm.bounds[0][self.inds_free]
            x0[lower_bound_violation] = \
                self.fm.bounds[0][self.inds_free][lower_bound_violation] + eps

            upper_bound_violation = x0 > self.fm.bounds[1][self.inds_free]
            x0[upper_bound_violation] = \
                self.fm.bounds[1][self.inds_free][upper_bound_violation] - eps
            del lower_bound_violation, upper_bound_violation

            # Find the full state vector with bounds checked
            x = self.full_statevector(x0)

            # Regardless of anything we did for the heuristic guess, bring the
            # static preseed back into play (only does anything if inds_preseed
            # is not blank)
            if len(self.inds_preseed) > 0:
                x0[self.inds_preseed] = combo

            # Record initializaation state
            geom.x_surf_init = x[self.fm.idx_surface]
            geom.x_RT_init = x[self.fm.idx_RT]

            # Seps is the covariance of "observation noise" including both
            # measurement noise from the instrument as well as variability due to
            # unknown variables. For speed, we will calculate it just once based
            # on the initial solution (a potential minor source of inaccuracy).
            Seps_inv, Seps_inv_sqrt = self.calc_Seps(x, meas, geom)

            def jac(x_free):
                """Short wrapper function for use with scipy opt"""
                return self.jacobian(x_free, geom, Seps_inv_sqrt)

            def err(x_free):
                """Short wrapper function for use with scipy opt and logging"""
                residual, x = self.loss_function(x_free, geom, Seps_inv_sqrt, meas)

                trajectory.append(x)

                it = len(trajectory)
                rs = float(np.sum(np.power(residual, 2)))
                sm = self.fm.summarize(x, geom)
                logging.debug('Iteration: %02i  Residual: %12.2f %s' %
                              (it, rs, sm))

                return np.real(residual)

            # Initialize and invert
            try:
                xopt = least_squares(err, x0, jac=jac, **self.least_squares_params)
                x_full_solution = self.full_statevector(xopt.x)
                trajectory.append(x_full_solution)
                solutions.append(trajectory)
                costs.append(np.sqrt(np.power(xopt.fun, 2).sum()))
            except scipy.linalg.LinAlgError:
                logging.warning('Optimization failed to converge')
                solutions.append(trajectory)
                costs.append(9e99)

        final_solution = np.array(solutions[np.argmin(costs)])
        return final_solution


    def forward_uncertainty(self, x, meas, geom):
        """
        Args:
            x: statevector
            meas: a one-D scipy vector of radiance in uW/nm/sr/cm2
            geom: a geometry object
        Returns:
            lamb: the converged lambertian surface reflectance
            path: the converged path radiance estimate
            mdl: the modeled radiance estimate
            S_hat: the posterior covariance of the state vector
            K: the derivative matrix d meas_x / d state_x
            G: the G matrix from the CD Rodgers 2000 formalism
        """

        dark_surface = np.zeros(self.fm.surface.wl.shape)
        path = self.fm.calc_meas(x, geom, rfl=dark_surface)
        mdl = self.fm.calc_meas(x, geom, rfl=None, Ls=None)
        lamb = self.fm.calc_lamb(x, geom)
        S_hat, K, G = self.calc_posterior(x, geom, meas)
        return lamb, mdl, path, S_hat, K, G



#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

from os.path import split, abspath
import numpy as np

from isofit.core.common import load_spectrum, expand_path, json_load_ascii
from isofit.core.instrument import Instrument
from isofit.core.geometry import Geometry


def generate_noise(config):
    """Add noise to a radiance spectrum or image."""

    config = json_load_ascii(config, shell_replace=True)
    configdir, configfile = split(abspath(config))

    infile = expand_path(configdir, config['input_radiance_file'])
    outfile = expand_path(configdir, config['output_radiance_file'])
    instrument = Instrument(config['instrument_model'])
    geom = Geometry()

    if infile.endswith('txt'):
        rdn, wl = load_spectrum(infile)
        Sy = instrument.Sy(rdn, geom)
        rdn_noise = rdn + np.random.multivariate_normal(np.zeros(rdn.shape), Sy)
        with open(outfile, 'w') as fout:
            for w, r in zip(wl, rdn_noise):
                fout.write('%8.5f %8.5f' % (w, r))
    else:
        raise ValueError("Image cubes not yet implemented.")



#! /usr/bin/env python3
#
#  Copyright 2019 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
from spectral.io import envi


def remap(inputfile, labels, outputfile, flag, chunksize):
    """."""

    ref_file = inputfile
    lbl_file = labels
    out_file = outputfile
    nchunk = chunksize

    ref_img = envi.open(ref_file+'.hdr', ref_file)
    ref_meta = ref_img.metadata
    ref_mm = ref_img.open_memmap(interleave='source', writable=False)
    ref = np.array(ref_mm[:, :])

    lbl_img = envi.open(lbl_file+'.hdr', lbl_file)
    lbl_meta = lbl_img.metadata
    labels = lbl_img.read_band(0)

    nl = int(lbl_meta['lines'])
    ns = int(lbl_meta['samples'])
    nb = int(ref_meta['bands'])

    out_meta = dict([(k, v) for k, v in ref_meta.items()])

    out_meta["samples"] = ns
    out_meta["bands"] = nb
    out_meta["lines"] = nl
    out_meta['data type'] = ref_meta['data type']
    out_meta["interleave"] = "bil"

    out_img = envi.create_image(out_file+'.hdr',  metadata=out_meta,
                                ext='', force=True)
    out_mm = out_img.open_memmap(interleave='source', writable=True)

    # Iterate through image "chunks," restoring as we go
    for lstart in np.arange(0, nl, nchunk):
        print(lstart)
        del out_mm
        out_mm = out_img.open_memmap(interleave='source', writable=True)

        # Which labels will we extract? ignore zero index
        lend = min(lstart+nchunk, nl)

        lbl = labels[lstart:lend, :]
        out = flag * np.ones((lbl.shape[0], nb, lbl.shape[1]))
        for row in range(lbl.shape[0]):
            for col in range(lbl.shape[1]):
                out[row, :, col] = np.squeeze(ref[int(lbl[row, col]), :])

        out_mm[lstart:lend, :, :] = out



#! /usr/bin/env python3
#
#  Copyright 2018 California Institute of Technology
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# ISOFIT: Imaging Spectrometer Optimal FITting
# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
#

import numpy as np
from scipy.linalg import block_diag, norm
from scipy.io import loadmat

from ..core.common import svd_inv
from .surface import Surface
from isofit.configs import Config


class MultiComponentSurface(Surface):
    """A model of the surface based on a collection of multivariate 
    Gaussians, with one or more equiprobable components and full 
    covariance matrices. 

    To evaluate the probability of a new spectrum, we calculate the
    Mahalanobis distance to each component cluster, and use that as our
    Multivariate Gaussian surface model.
    """

    def __init__(self, full_config: Config):
        """."""

        super().__init__(full_config)

        config = full_config.forward_model.surface

        # Models are stored as dictionaries in .mat format
        # TODO: inforce surface_file existence in the case of multicomponent_surface
        model_dict = loadmat(config.surface_file)
        self.components = list(zip(model_dict['means'], model_dict['covs']))
        self.n_comp = len(self.components)
        self.wl = model_dict['wl'][0]
        self.n_wl = len(self.wl)

        # Set up normalization method
        self.normalize = model_dict['normalize']
        if self.normalize == 'Euclidean':
            self.norm = lambda r: norm(r)
        elif self.normalize == 'RMS':
            self.norm = lambda r: np.sqrt(np.mean(pow(r, 2)))
        elif self.normalize == 'None':
            self.norm = lambda r: 1.0
        else:
            raise ValueError('Unrecognized Normalization: %s\n' %
                             self.normalize)

        self.selection_metric = config.selection_metric
        self.select_on_init = config.select_on_init

        # Reference values are used for normalizing the reflectances.
        # in the VSWIR regime, reflectances are normalized so that the model
        # is agnostic to absolute magnitude.
        self.refwl = np.squeeze(model_dict['refwl'])
        self.idx_ref = [np.argmin(abs(self.wl-w))
                        for w in np.squeeze(self.refwl)]
        self.idx_ref = np.array(self.idx_ref)

        # Cache some important computations
        self.Covs, self.Cinvs, self.mus = [], [], []
        for i in range(self.n_comp):
            Cov = self.components[i][1]
            self.Covs.append(np.array([Cov[j, self.idx_ref]
                                       for j in self.idx_ref]))
            self.Cinvs.append(svd_inv(self.Covs[-1]))
            self.mus.append(self.components[i][0][self.idx_ref])

        # Variables retrieved: each channel maps to a reflectance model parameter
        rmin, rmax = 0, 2.0
        self.statevec_names = ['RFL_%04i' % int(w) for w in self.wl]
        self.bounds = [[rmin, rmax] for w in self.wl]
        self.scale = [1.0 for w in self.wl]
        self.init = [0.15 * (rmax-rmin)+rmin for v in self.wl]
        self.idx_lamb = np.arange(self.n_wl)
        self.n_state = len(self.statevec_names)

    def component(self, x, geom):
        """We pick a surface model component using the Mahalanobis distance.

        This always uses the Lambertian (non-specular) version of the 
        surface reflectance. If the forward model initialize via heuristic
        (i.e. algebraic inversion), the component is only calculated once
        based on that first solution. That state is preserved in the 
        geometry object.
        """

        if self.n_comp <= 1:
            return 0
        elif hasattr(geom, 'surf_cmp_init'):
            return geom.surf_cmp_init
        elif self.select_on_init and hasattr(geom, 'x_surf_init'):
            x_surface = geom.x_surf_init
        else:
            x_surface = x

        # Get the (possibly normalized) reflectance
        lamb = self.calc_lamb(x_surface, geom)
        lamb_ref = lamb[self.idx_ref]
        lamb_ref = lamb_ref / self.norm(lamb_ref)

        # Mahalanobis or Euclidean distances
        mds = []
        for ci in range(self.n_comp):
            ref_mu = self.mus[ci]
            ref_Cinv = self.Cinvs[ci]
            if self.selection_metric == 'Mahalanobis':
                md = (lamb_ref - ref_mu).T.dot(ref_Cinv).dot(lamb_ref - ref_mu)
            else:
                md = sum(pow(lamb_ref - ref_mu, 2))
            mds.append(md)
        closest = np.argmin(mds)

        if self.select_on_init and hasattr(geom, 'x_surf_init') and \
                (not hasattr(geom, 'surf_cmp_init')):
            geom.surf_cmp_init = closest
        return closest

    def xa(self, x_surface, geom):
        """Mean of prior distribution, calculated at state x. We find
        the covariance in a normalized space (normalizing by z) and then un-
        normalize the result for the calling function. This always uses the
        Lambertian (non-specular) version of the surface reflectance."""

        lamb = self.calc_lamb(x_surface, geom)
        lamb_ref = lamb[self.idx_ref]
        mu = np.zeros(self.n_state)
        ci = self.component(x_surface, geom)
        lamb_mu = self.components[ci][0]
        lamb_mu = lamb_mu * self.norm(lamb_ref)
        mu[self.idx_lamb] = lamb_mu
        return mu

    def Sa(self, x_surface, geom):
        """Covariance of prior distribution, calculated at state x. We find
        the covariance in a normalized space (normalizing by z) and then un-
        normalize the result for the calling function."""

        lamb = self.calc_lamb(x_surface, geom)
        lamb_ref = lamb[self.idx_ref]
        ci = self.component(x_surface, geom)
        Cov = self.components[ci][1]
        Cov = Cov * (self.norm(lamb_ref)**2)

        # If there are no other state vector elements, we're done.
        if len(self.statevec_names) == len(self.idx_lamb):
            return Cov

        # Embed into a larger state vector covariance matrix
        nprefix = self.idx_lamb[0]
        nsuffix = len(self.statevec_names) - self.idx_lamb[-1] - 1
        Cov_prefix = np.zeros((nprefix, nprefix))
        Cov_suffix = np.zeros((nsuffix, nsuffix))
        return block_diag(Cov_prefix, Cov, Cov_suffix)

    def fit_params(self, rfl_meas, geom, *args):
        """Given a reflectance estimate, fit a state vector."""

        x_surface = np.zeros(len(self.statevec_names))
        if len(rfl_meas) != len(self.idx_lamb):
            raise ValueError('Mismatched reflectances')
        for i, r in zip(self.idx_lamb, rfl_meas):
            x_surface[i] = max(self.bounds[i][0]+0.001,
                               min(self.bounds[i][1]-0.001, r))
        return x_surface

    def calc_rfl(self, x_surface, geom):
        """Non-Lambertian reflectance."""

        return self.calc_lamb(x_surface, geom)

    def calc_lamb(self, x_surface, geom):
        """Lambertian reflectance."""

        return x_surface[self.idx_lamb]

    def drfl_dsurface(self, x_surface, geom):
        """Partial derivative of reflectance with respect to state vector, 
        calculated at x_surface."""

        return self.dlamb_dsurface(x_surface, geom)

    def dlamb_dsurface(self, x_surface, geom):
        """Partial derivative of Lambertian reflectance with respect to 
        state vector, calculated at x_surface."""

        dlamb = np.eye(self.n_wl, dtype=float)
        nprefix = self.idx_lamb[0]
        nsuffix = self.n_state - self.idx_lamb[-1] - 1
        prefix = np.zeros((self.n_wl, nprefix))
        suffix = np.zeros((self.n_wl, nsuffix))
        return np.concatenate((prefix, dlamb, suffix), axis=1)

    def calc_Ls(self, x_surface, geom):
        """Emission of surface, as a radiance."""

        return np.zeros(self.n_wl, dtype=float)

    def dLs_dsurface(self, x_surface, geom):
        """Partial derivative of surface emission with respect to state vector, 
        calculated at x_surface."""

        dLs = np.zeros((self.n_wl, self.n_wl), dtype=float)
        nprefix = self.idx_lamb[0]
        nsuffix = len(self.statevec_names) - self.idx_lamb[-1] - 1
        prefix = np.zeros((self.n_wl, nprefix))
        suffix = np.zeros((self.n_wl, nsuffix))
        return np.concatenate((prefix, dLs, suffix), axis=1)

    def summarize(self, x_surface, geom):
        """Summary of state vector."""

        if len(x_surface) < 1:
            return ''
        return 'Component: %i' % self.component(x_surface, geom)



# David R Thompson, Adam Erickson

from isofit.core.isofit import Isofit
from isofit.utils import surface_model

# Build the surface model
surface_model("configs/ang20171108t184227_surface.json")

# Run retrievals
model = Isofit("configs/ang20171108t173546_darklot.json")
model.run()
del model

model = Isofit("configs/ang20171108t173546_horse.json")
model.run()
del model

model = Isofit("configs/ang20171108t184227_astrored.json")
model.run()
del model

model = Isofit("configs/ang20171108t184227_astrogreen.json")
model.run()
del model

model = Isofit("configs/ang20171108t184227_beckmanlawn.json")
model.run()
del model

model = Isofit("configs/ang20171108t184227_beckmanlawn-oversmoothed.json")
model.run()
del model

model = Isofit("configs/ang20171108t184227_beckmanlawn-undersmoothed.json")
model.run()
del model
